from typing import Tuple

from dexp.processing.backends.backend import Backend


def add_border(image,
               width: int = 2,
               color: Tuple[float, float, float, float] = None,
               over_image: bool = False,
               rgba_value_max: float = 255
               ):
    """
    Adds a color border to an image

    Parameters
    ----------
    image: Base image.
    width: Width of border.
    color: Border color as tuple (R, G, B, A) of normalised floats.
    over_image: If True the border is not added but overlayed over the image, the image does not change size.
    rgba_value_max: max value for rgba values.

    Returns
    -------
    Image with border added or overlayed

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # If border is 0 then rwe return the image unchanged:
    if width == 0:
        return image

    # Move to backend:
    image = Backend.to_backend(image)

    # Default color:
    if color is None:
        color = (1, 1, 1, 1)

    # Bring alpha to the front:
    image = image.transpose(2, 0, 1)

    shape = image.shape

    # New image with border:
    new_shape = shape if over_image else (shape[0],) + tuple(s + 2 * width for s in shape[1:])

    image_with_border = xp.zeros(shape=new_shape, dtype=image.dtype)

    for i, channel in enumerate(image):
        value = int(color[i] * rgba_value_max)
        if over_image:
            channel = channel[..., width:-width, width:-width]

        padded_channel = xp.pad(channel,
                                pad_width=width,
                                mode='constant',
                                constant_values=value)
        image_with_border[i,...] = padded_channel

    image_with_border = image_with_border.transpose(1, 2, 0)

    return image_with_border



from dexp.processing.backends.backend import Backend


def fit_to_shape(array, shape):
    """
    Pads or crops an array to attain a certain shape

    Parameters
    ----------
    backend : backend to use
    array : array to pad
    shape : shape to attain by cropping or padding each dimension

    Returns
    -------
    Array of requested shape

    """

    length_diff = tuple(u - v for u, v in zip(shape, array.shape))

    if any(x < 0 for x in length_diff):
        # we need to crop at least one dimension:
        slicing = tuple(slice(0, s) for s in shape)
        array = array[slicing]

    # Independently of whether we had to crop a dimension, we proceed with eventual padding:
    length_diff = tuple(u - v for u, v in zip(shape, array.shape))

    if any(x > 0 for x in length_diff):
        xp = Backend.get_xp_module()
        pad_width = tuple(tuple((0, d)) for d in length_diff)
        array = xp.pad(array, pad_width=pad_width)

    return array



from typing import Union, Tuple, Sequence

from dexp.processing.backends.backend import Backend
from dexp.processing.color.blend import blend_color_images
from dexp.processing.color.cairo_utils import get_array_for_cairo_surface


def insert_scale_bar(image,
                     length_in_unit: float = 1,
                     pixel_scale: float = 1,
                     bar_height: int = 4,
                     margin: float = 1,
                     translation: Union[str, Tuple[Union[int, float], ...]] = 'bottom_right',
                     color: Tuple[float, float, float, float] = None,
                     number_format: str = '{:.1f}',
                     font_name: str = "Helvetica",
                     font_size: float = 32,
                     unit: str = 'μm',
                     mode: str = 'max'):
    """
    Inserts a scale bar into an image.

    Parameters
    ----------
    image: Image into which to insert the scale bar.
    length_in_unit: Length of scale bar in the provided unit.
    pixel_scale: conversion factor from pixels to units -- what is the side length of a pixel/voxel in units.
    bar_height: Height of th scale bar in pixels
    margin: margin around bar expressed in units relative to the text height
    translation: Positions of the scale bar in pixels in numpy order: (y, x). Can also be a string: 'bottom_left', 'bottom_right', 'top_left', 'top_right'.
    color: Color of the bar and text as tuple of 4 values: (R, G, B, A)
    number_format: Format string to represent the start and end values.
    font_name: Font name.
    font_size: Font size in pixels.
    unit: Unit name.
    mode: Blending mode. See function 'blend_color_images' for available blending modes.

    Returns
    -------
    Image with inserted scale bar.

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # Move to backend:
    image = Backend.to_backend(image)

    if color is None:
        color = (1, 1, 1, 1)

    # Image shape:
    height, width = image.shape[:-1]

    # Bar length in pixels/voxels:
    bar_length = length_in_unit / pixel_scale

    # Replace um with µm:
    if unit.strip() == 'um':
        unit = 'µm'

    scale_bar_array = _generate_scale_bar_image(width,
                                                height,
                                                translation,
                                                bar_height,
                                                bar_length,
                                                margin,
                                                color,
                                                font_name,
                                                font_size,
                                                length_in_unit,
                                                number_format,
                                                unit)

    # Blend images:
    result = blend_color_images(images=(image, scale_bar_array),
                                alphas=(1, 1),
                                modes=('max', mode))

    return result, scale_bar_array


def _generate_scale_bar_image(width,
                              height,
                              translation,
                              bar_height,
                              bar_length,
                              margin,
                              color,
                              font_name,
                              font_size,
                              length_in_unit,
                              number_format,
                              unit):
    # Create surface:
    import cairocffi as cairo
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    context = cairo.Context(surface)
    context.scale(1, 1)

    # Configure text rendering:
    context.select_font_face(font_name,
                             cairo.FONT_SLANT_NORMAL,
                             cairo.FONT_WEIGHT_NORMAL)
    context.set_font_size(font_size)

    # text to be rendered:
    text = f"{number_format.format(length_in_unit)} {unit}"

    # determine text height and width:
    text_width = context.text_extents(text)[2]
    text_height = context.text_extents(text)[3]

    # Margins:
    margin_height = margin * text_height
    margin_width = margin * text_height

    # First we turn off antialiasing, works better:
    context.set_antialias(cairo.ANTIALIAS_NONE)

    # Turn off alpha blending and clears the image:
    context.set_operator(cairo.OPERATOR_SOURCE)
    context.set_source_rgba(0, 0, 0, 0)
    context.rectangle(0, 0, 1, 1)
    context.fill()

    # We set the color for the bar:
    context.set_source_rgba(*color)

    # Set the position of the bar, that is the left-most end of it, including the margin:
    if type(translation) == str:
        if 'top' in translation:
            y = margin_height
        elif 'bottom' in translation:
            y = height - (margin_height + bar_height)
        if 'left' in translation:
            x = margin_width
        elif 'right' in translation:
            x = width - (margin_width + bar_length / 2 + max(bar_length / 2, text_width / 2))
    elif type(translation) == tuple:
        y, x = translation
    else:
        raise ValueError(f"Invalid translation: {translation}")


    # We draw the scale bar itself:
    context.rectangle(x, y - bar_height / 2, bar_length, bar_height)
    context.fill()

    # Turn on antialiasing again for text:
    context.set_antialias(cairo.ANTIALIAS_SUBPIXEL)

    # Turn back on alpha blending:
    context.set_operator(cairo.OPERATOR_OVER)

    # Set text color:
    context.set_source_rgba(*color)

    # draw text at correct location:
    if 'top' in translation:
        context.move_to(x + bar_length / 2 - text_width / 2, y + (3 * text_height / 2))
    else:
        context.move_to(x + bar_length / 2 - text_width / 2, y - text_height / 2)
    context.show_text(text)

    # Get array from surface:
    surface_array = get_array_for_cairo_surface(surface)
    return surface_array



from napari import gui_qt, Viewer

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.clean_dark_regions import clean_dark_regions
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_clean_dark_regions_numpy():
    with NumpyBackend():
        demo_clean_dark_regions_data()


def demo_clean_dark_regions_cupy():
    # try:
    with CupyBackend():
        demo_clean_dark_regions_data()
    # except (ModuleNotFoundError, NotImplementedError):
    #    print("Cupy module not found! ignored!")


def demo_clean_dark_regions_data(length_xy=256):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=True,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=1,
                                                                      independent_haze=False,
                                                                      background_stength=0.1)

    # remove zero level
    image = xp.clip(image - 95, 0, None)

    with timeit('dehaze_new'):
        cleaned = clean_dark_regions(image, size=5, threshold=30, in_place=False)

    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt')
        viewer.add_image(_c(background), name='background')
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(cleaned), name='cleaned', gamma=0.1)
        viewer.grid.enabled = True


if __name__ == "__main__":
    demo_clean_dark_regions_cupy()
    demo_clean_dark_regions_numpy()



from os import makedirs
from os.path import exists, join
from typing import Any, Tuple, Sequence

import imageio
from arbol.arbol import aprint, asection
from joblib import delayed, Parallel

from dexp.datasets.base_dataset import BaseDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.color.projection import project_image


def dataset_projection_rendering(input_dataset: BaseDataset,
                                 output_path: str = None,
                                 channels: Sequence[str] = None,
                                 slicing: Any = None,
                                 overwrite: bool = False,
                                 axis: int = 0,
                                 dir: int = -1,
                                 mode: str = 'colormax',
                                 clim: Tuple[float, float] = None,
                                 attenuation: float = 0.05,
                                 gamma: float = 1.0,
                                 dlim: Tuple[float, float] = None,
                                 colormap: str = None,
                                 rgbgamma: float = 1.0,
                                 transparency: bool = False,
                                 legendsize: float = 1.0,
                                 legendscale: float = 1.0,
                                 legendtitle: str = 'color-coded depth (voxels)',
                                 legendtitlecolor: Tuple[float, float, float, float] = (1, 1, 1, 1),
                                 legendposition: str = 'bottom_left',
                                 legendalpha: float = 1.0,
                                 step: int = 1,
                                 workers: int = -1,
                                 workersbackend: str = 'threading',
                                 devices: Tuple[int, ...] = (0,),
                                 stop_at_exception=True):
    for channel in channels:

        # Ensures that the output folder exists per channel:
        if len(channels) == 1:
            channel_output_path = output_path
        else:
            channel_output_path = output_path + f'_{channel}'

        makedirs(channel_output_path, exist_ok=True)

        with asection(f"Channel '{channel}' shape: {input_dataset.shape(channel)}:"):
            aprint(input_dataset.info(channel))

        array = input_dataset.get_array(channel, wrap_with_dask=True)

        if slicing:
            array = array[slicing]

        aprint(f"Rendering array of shape={array.shape} and dtype={array.dtype} for channel '{channel}'.")

        nbframes = array.shape[0]

        with asection("Rendering:"):

            def process(tp, _clim, device):
                try:
                    with asection(f"Rendering Frame     : {tp:05}"):

                        filename = join(channel_output_path, f"frame_{tp:05}.png")

                        if overwrite or not exists(filename):

                            with asection("Loading stack..."):
                                stack = array[tp].compute()

                            with BestBackend(device, exclusive=True, enable_unified_memory=True):
                                if _clim is not None:
                                    aprint(f"Using provided min and max for contrast limits: {_clim}")
                                    min_value, max_value = (float(strvalue) for strvalue in _clim.split(','))
                                    _clim = (min_value, max_value)

                                with asection(f"Projecting image of shape: {stack.shape} "):
                                    projection = project_image(stack,
                                                               axis=axis,
                                                               dir=dir,
                                                               mode=mode,
                                                               attenuation=attenuation,
                                                               attenuation_min_density=0.002,
                                                               attenuation_filtering=4,
                                                               gamma=gamma,
                                                               clim=_clim,
                                                               cmap=colormap,
                                                               dlim=dlim,
                                                               rgb_gamma=rgbgamma,
                                                               transparency=transparency,
                                                               legend_size=legendsize,
                                                               legend_scale=legendscale,
                                                               legend_title=legendtitle,
                                                               legend_title_color=legendtitlecolor,
                                                               legend_position=legendposition,
                                                               legend_alpha=legendalpha)

                                with asection(f"Saving frame {tp} as: {filename}"):
                                    imageio.imwrite(filename,
                                                    Backend.to_numpy(projection),
                                                    compress_level=1)

                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while processing time point {tp} !")
                    import traceback
                    traceback.print_exc()

                    if stop_at_exception:
                        raise error

            if workers == -1:
                workers = len(devices)
            aprint(f"Number of workers: {workers}")

            if workers > 1:
                Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp, clim, devices[tp % len(devices)]) for tp in range(0, nbframes, step))
            else:
                for tp in range(0, nbframes, step):
                    process(tp, clim, devices[0])



from os import system
from os.path import exists, join

import click
from arbol.arbol import aprint, asection

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group()
def cli():
    aprint("__________________________________________")
    aprint("  DEXP -- Data EXploration & Processing   ")
    aprint("  System installation commands            ")
    aprint("  Royer lab                               ")
    aprint("__________________________________________")
    aprint("")
    # input("Press Enter to continue...")


@click.command()
@click.argument('cuda', nargs=1)
def cudalibs(cuda: str):
    """ Install additional CUDA libraries.

    CUDA is the cuda version: 11.2, 11.1, 11.0, ...
    """
    with asection(f"Install additional CUDA libraries"):

        from os.path import expanduser
        home = expanduser("~")
        if not exists(join(home, f'.cupy/cuda_lib/{cuda}')):
            aprint(f'Installing CUDNN for CUDA {cuda}')
            system(f'python -m cupyx.tools.install_library --library cudnn --cuda {cuda}')
            aprint(f'Installing CUTENSOR for CUDA {cuda}')
            system(f'python -m cupyx.tools.install_library --library cutensor --cuda {cuda}')
        else:
            aprint(f'Libraries already installed')



cli.add_command(cudalibs)



import numpy
from numpy.linalg import norm
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.lipshitz_correction import lipschitz_continuity_correction
from dexp.utils.timeit import timeit


def test_lipschitz_continuity_correction_numpy():
    with NumpyBackend():
        _test_lipschitz_continuity_correction()


def test_lipschitz_continuity_correction_cupy():
    try:
        with CupyBackend():
            _test_lipschitz_continuity_correction()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def _test_lipschitz_continuity_correction(length_xy=128):
    image = camera().astype(numpy.float32) / 255
    noisy = random_noise(image.copy(), mode="gaussian", var=0.005, seed=0, clip=False)
    noisy = random_noise(noisy, mode="s&p", amount=0.03, seed=0, clip=False)

    with timeit('lipschitz_continuity_correction'):
        corrected = lipschitz_continuity_correction(image, lipschitz=0.15, in_place=False)
        corrected = Backend.to_numpy(corrected)

    assert corrected is not image
    assert corrected.shape == image.shape
    assert corrected.dtype == image.dtype

    error0 = norm(noisy - image, ord=1) / image.size
    error = norm(corrected - image, ord=1) / image.size

    print(f"Error noisy = {error0}")
    print(f"Error = {error}")

    assert error < 0.001
    assert error0 > 4 * error



import numpy
from dexp.enhance import sharpen
from napari import gui_qt, Viewer
from scipy.ndimage import zoom
from tifffile import imread, imwrite

from dexp.processing.isonet import IsoNet
from dexp.utils.timeit import timeit

# image = imread('data/retina/cropped_farred_RFP_GFP_2109175_2color_sub_10.20.tif')
image = imread('../data/zfish/zfish2.tif')

image = zoom(image, zoom=(1, 0.5, 0.5), order=0)

image = sharpen(image, mode='hybrid', min=0, max=1024, margin_pad=False)

print(f"shape={image.shape}, dtype={image.dtype}, min={image.min()}, max={image.max()}")
image_crop = image  # [36-16:36+16,750-128:750+128,750-128:750+128]
print(image_crop.shape)

dxy = 0.4 * 2
dz = 6.349
subsampling = dz / dxy
print(f'subsampling={subsampling}')

# psf = nikon16x08na(dxy=dxy,dz=dz, size=16)[:,(15-1)//2,:]
psf = numpy.ones((3, 3)) / 9

isonet = IsoNet()

# with timeit("Preparation:"):
#     isonet.prepare(image_crop, subsampling=subsampling, psf=psf, threshold=0.97)
#
# with timeit("Training:"):
#     isonet.train(max_epochs=40)

with timeit("Evaluation:"):
    restored = isonet.apply_pair(image_crop, subsampling=subsampling, batch_size=1)

print(restored.shape)

imwrite('../data/isotropic.tif', restored, imagej=True, resolution=(1 / dxy, 1 / dxy), metadata={'spacing': dz, 'unit': 'um'})

image = zoom(image, zoom=(subsampling, 1, 1), order=0)

with gui_qt():
    viewer = Viewer()

    viewer.add_image(image, name='image (all)')
    viewer.add_image(image_crop, name='image selection')
    viewer.add_image(restored, name='restored')
    viewer.add_image(psf, name='psf')



from typing import Tuple

import numpy

from dexp.processing.backends.backend import Backend


def create_cuda_texture(array,
                        texture_shape: Tuple[int, ...] = None,
                        num_channels: int = 1,
                        normalised_values: bool = False,
                        normalised_coords: bool = False,
                        sampling_mode: str = 'linear',
                        address_mode: str = 'clamp',
                        dtype=None):
    """ Creates a Cuda texture and takes care of a lot of the needed 'know-how'

    Parameters
    ----------
    array
    texture_shape
    num_channels
    normalised_values
    normalised_coords
    sampling_mode
    address_mode
    dtype

    Returns
    -------

    """
    import cupy

    if texture_shape is None:
        if num_channels > 1:
            texture_shape = array.shape[0:-1]
        else:
            texture_shape = array.shape

    if array.dtype == numpy.float16 or dtype == numpy.float16:
        raise ValueError('float16 types not yet supported!')

    if dtype is None:
        dtype = array.dtype

    if not (1 <= len(texture_shape) <= 3):
        raise ValueError(f"Invalid number of dimensions ({len(texture_shape)}), must be 1, 2 or 3 (shape={texture_shape}) ")

    if not (num_channels == 1 or num_channels == 2 or num_channels == 4):
        raise ValueError(f"Invalid number of channels ({num_channels}), must be 1, 2., 3 or 4")

    if array.size != numpy.prod(texture_shape) * num_channels:
        raise ValueError(f"Texture shape {texture_shape}, num of channels ({num_channels}), and array size ({array.size}) are mismatched!")

    dtype = numpy.dtype(dtype)

    if array.dtype != dtype:
        array = array.astype(dtype, copy=False)

    nbits = 8 * dtype.itemsize
    channels = (nbits,) * num_channels + (0,) * (4 - num_channels)
    if 'f' in dtype.kind:
        channel_type = cupy.cuda.runtime.cudaChannelFormatKindFloat
    elif 'i' in dtype.kind:
        channel_type = cupy.cuda.runtime.cudaChannelFormatKindSigned
    elif 'u' in dtype.kind:
        channel_type = cupy.cuda.runtime.cudaChannelFormatKindUnsigned
    else:
        raise ValueError(f"dtype '{address_mode}' is not supported")

    format_descriptor = cupy.cuda.texture.ChannelFormatDescriptor(*channels, channel_type)

    cuda_array = cupy.cuda.texture.CUDAarray(format_descriptor, *(texture_shape[::-1]))
    ressource_descriptor = cupy.cuda.texture.ResourceDescriptor(cupy.cuda.runtime.cudaResourceTypeArray, cuArr=cuda_array)

    if address_mode == 'clamp':
        address_mode = cupy.cuda.runtime.cudaAddressModeClamp
    elif address_mode == 'border':
        address_mode = cupy.cuda.runtime.cudaAddressModeBorder
    elif address_mode == 'wrap':
        address_mode = cupy.cuda.runtime.cudaAddressModeWrap
    elif address_mode == 'mirror':
        address_mode = cupy.cuda.runtime.cudaAddressModeMirror
    else:
        raise ValueError(f"Address mode '{address_mode}' is not supported")

    address_mode = (address_mode,) * len(texture_shape)

    if sampling_mode == 'nearest':
        filter_mode = cupy.cuda.runtime.cudaFilterModePoint
    elif sampling_mode == 'linear':
        filter_mode = cupy.cuda.runtime.cudaFilterModeLinear
    else:
        raise ValueError(f"Sampling mode '{sampling_mode}' is not supported")

    if normalised_values:
        read_mode = cupy.cuda.runtime.cudaReadModeNormalizedFloat
    else:
        read_mode = cupy.cuda.runtime.cudaReadModeElementType

    texture_descriptor = cupy.cuda.texture.TextureDescriptor(addressModes=address_mode,
                                                             filterMode=filter_mode,
                                                             readMode=read_mode,
                                                             sRGB=None,
                                                             borderColors=None,
                                                             normalizedCoords=normalised_coords,
                                                             maxAnisotropy=None)

    texture_object = cupy.cuda.texture.TextureObject(ressource_descriptor,
                                                     texture_descriptor)

    # 'copy_from' from CUDAArray requires that the num of channels be multiplied to the last axis of the array (see cupy docs!)
    if num_channels > 1:
        array_shape_for_copy = texture_shape[:-1] + (texture_shape[-1] * num_channels,)
    else:
        array_shape_for_copy = texture_shape
    axp = cupy.get_array_module(array)
    array = axp.reshape(array, newshape=array_shape_for_copy)

    if not array.flags.owndata or not not array.flags.c_contiguous:
        # the array must be contiguous, we check if this is a derived array,
        # if yes we must unfortunately copy the data...
        array = array.copy()

    # We need to synchronise otherwise some weird stuff happens! see warp 3d demo does not work withoutv this!
    # Backend.current().synchronise()
    cuda_array.copy_from(array, stream=Backend.current().stream)
    # Backend.current().synchronise()

    del format_descriptor, texture_descriptor, ressource_descriptor

    return texture_object, cuda_array



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.clean_dark_regions import clean_dark_regions
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def test_clean_dark_regions_numpy():
    with NumpyBackend():
        _test_clean_dark_regions()


def test_clean_dark_regions_cupy():
    try:
        with CupyBackend():
            _test_clean_dark_regions()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def _test_clean_dark_regions(length_xy=128):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=True,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=1,
                                                                      independent_haze=False,
                                                                      background_stength=0.05,
                                                                      dtype=numpy.float32)

    # remove zero level
    image = xp.clip(image - 95, 0, None)

    with timeit('clean_dark_regions'):
        cleaned = clean_dark_regions(image, size=3, threshold=30, in_place=False)

    assert cleaned is not image
    assert cleaned.shape == image.shape
    assert cleaned.dtype == image.dtype

    # compute percentage of haze removed:
    background_voxels_image = (1 - image_gt) * image
    background_voxels_dehazed = (1 - image_gt) * cleaned
    total_haze = xp.sum(background_voxels_image)
    total_remaining_haze = xp.sum(background_voxels_dehazed)
    percent_removed = (total_haze - total_remaining_haze) / total_haze
    print(f"percent_removed = {percent_removed}")

    # compute error on non-background voxels
    non_background_voxels_image = image_gt * image
    non_background_voxels_dehazed = image_gt * cleaned
    average_error = xp.mean(xp.absolute(non_background_voxels_image - non_background_voxels_dehazed))
    print(f"average_error = {average_error}")

    assert percent_removed > 0.1
    assert percent_removed < 3000



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.fusion.dct_fusion import fuse_dct_nd
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data
from dexp.utils.timeit import timeit


def test_dct_fusion_numpy():
    with NumpyBackend():
        dct_fusion()


def dct_fusion(length_xy=128):
    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False,
                                                                                       length_xy=length_xy,
                                                                                       length_z_factor=4)
    with timeit("dcf fusion + data transfer"):
        image_fused = fuse_dct_nd(image1, image2)
        image_fused = Backend.to_numpy(image_fused)

    image_gt = Backend.to_numpy(image_gt)
    error = numpy.median(numpy.abs(image_gt - image_fused))
    print(f"error={error}")
    assert error < 22

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image_gt, name='image_gt')
    #     viewer.add_image(image_lowq, name='image_lowq')
    #     viewer.add_image(blend_a, name='blend_a')
    #     viewer.add_image(blend_b, name='blend_b')
    #     viewer.add_image(image1, name='image1')
    #     viewer.add_image(image2, name='image2')
    #     viewer.add_image(image_fused, name='image_fused')



import cupy

from dexp.processing.backends._cupy.texture.texture import create_cuda_texture
from dexp.processing.backends.backend import Backend


def _warp_1d_cupy(image,
                  vector_field,
                  mode,
                  block_size: int = 128):
    """

    Parameters
    ----------
    image
    vector_field
    mode
    block_size

    Returns
    -------

    """
    xp = Backend.get_xp_module()
    source = r'''
                extern "C"{
                __global__ void warp_1d(float* warped_image,
                                        cudaTextureObject_t input_image,
                                        cudaTextureObject_t vector_field,
                                        int width)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;

                    if (x < width)
                    {
                        // coordinates in coord-normalised vector_field texture:
                        float u = float(x)/width;

                        // Obtain linearly interpolated vector at (u,):
                        float vector = tex1D<float>(vector_field, u);

                        // Obtain the shifted coordinates of the source voxel: 
                        float sx = 0.5f + float(x) - vector;

                        // Sample source image for voxel value:
                        float value = tex1D<float>(input_image, sx);

                        //printf("(%f, %f)=%f\n", sx, sy, value);

                        // Store interpolated value:
                        warped_image[x] = value;
 
                    }
                }
                }
                '''

    if image.ndim != 1 or vector_field.ndim != 1:
        raise ValueError("image or vector field has wrong number of dimensions!")

    # set up textures:
    input_image_tex, input_image_cudarr = create_cuda_texture(image,
                                                              num_channels=1,
                                                              normalised_coords=False,
                                                              sampling_mode='linear',
                                                              address_mode=mode,
                                                              )

    vector_field_tex, vector_field_cudarr = create_cuda_texture(vector_field,
                                                                num_channels=1,
                                                                normalised_coords=True,
                                                                sampling_mode='linear',
                                                                address_mode='clamp')

    # Set up resulting image:
    warped_image = xp.empty(shape=image.shape, dtype=image.dtype)

    # get the kernel, which copies from texture memory
    warp_1d_kernel = cupy.RawKernel(source, 'warp_1d')

    # launch kernel
    width, = image.shape

    grid_x = (width + block_size - 1) // block_size
    warp_1d_kernel((grid_x,),
                   (block_size,),
                   (warped_image, input_image_tex, vector_field_tex, width))

    del input_image_tex, input_image_cudarr, vector_field_tex, vector_field_cudarr

    return warped_image



import numpy
from scipy.ndimage import median_filter
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.lipshitz_correction import lipschitz_continuity_correction


def demo_lipschitz_correction_numpy():
    with NumpyBackend():
        demo_lipschitz_correction()


def demo_lipschitz_correction_cupy():
    try:
        with CupyBackend():
            demo_lipschitz_correction()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_lipschitz_correction():
    image = camera().astype(numpy.float32) / 255
    image = random_noise(image, mode="gaussian", var=0.005, seed=0, clip=False)
    image = random_noise(image, mode="s&p", amount=0.03, seed=0, clip=False)

    corrected = lipschitz_continuity_correction(image, lipschitz=0.15, in_place=False)

    median = median_filter(image, size=3)

    import napari
    with napari.gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = napari.Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(median), name='median_filtered')
        viewer.add_image(_c(corrected), name='corrected')
        viewer.grid.enabled = True


if __name__ == "__main__":
    demo_lipschitz_correction_cupy()
    demo_lipschitz_correction_numpy()



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.utils.projection_generator import projection_generator


def test_projection_generator_numpy():
    with NumpyBackend():
        _test_projection_generator()


def test_projection_generator_cupy():
    try:
        with CupyBackend():
            _test_projection_generator()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_projection_generator():
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image = xp.random.rand(12, 14, 17, 19)

    projections = list(projection_generator(image))
    assert len(projections) == 6

    for u, v, projection in projection_generator(image):
        assert 0 <= u <= 3
        assert 0 <= v <= 3
        assert u < v

        assert projection.shape[0] == image.shape[u]
        assert projection.shape[1] == image.shape[v]



from typing import Tuple

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.color.cairo_utils import get_array_for_cairo_surface
from dexp.processing.color.colormap import rgb_colormap


def depth_color_scale_legend(cmap,
                             start: float,
                             end: float,
                             flip: bool = False,
                             number_format: str = '{:.1f}',
                             font_name: str = "Helvetica",
                             font_size: float = 32.0,
                             title: str = '',
                             color: Tuple[float, float, float, float] = (1, 1, 1, 1),
                             bar_height: float = 0.15,
                             size: float = 1,
                             pixel_resolution: int = 512):
    """
    Produces a color bar legend.

    Note: if you need to specify the unit as microns, use this symbol: μm
    
    Parameters
    ----------
    cmap: Color map to use
    flip: Set to rue to flip the colormap
    start: Start value
    end: End value
    number_format: Format string to represent the start and end values.
    font_name: Font name.
    font_size: Font size in pixels.
    title: Title for bar legend
    color: Title color as a tuple of normalised floats: (R, G, B, A)  (values between 0 and 1).
    bar_height: Height of the color bar in relative units [0,1].
    size: Overall size factor (default: 1)
    pixel_resolution: Base resolution in pixels to color at (pixels per unit of scale).

    Returns
    -------

    """
    width = int(size * pixel_resolution)
    height = int(size * pixel_resolution)

    # First we build the depth ramp:
    depth_ramp = numpy.linspace(0, 1, num=255)
    depth_ramp = numpy.flip(depth_ramp) if flip else depth_ramp

    # get color ramp:
    color_ramp = rgb_colormap(depth_ramp, cmap=cmap, bytes=False)

    # Create a Cairo surface:
    import cairocffi as cairo
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    context = cairo.Context(surface)
    context.scale(width, height)

    # Adjust font size to relative coordinates:
    font_size /= pixel_resolution

    # we draw the depth ramp:
    context.set_antialias(cairo.ANTIALIAS_NONE)
    delta = 1.0 / 255
    for i in range(255):
        # determine color:
        _color = color_ramp[i]
        _color = (float(_color[0]), float(_color[1]), float(_color[2]), 1.0)
        context.set_source_rgba(*_color)

        # set position:
        pos_x = i * delta
        pos_y = 0

        # draw rectangle:
        context.rectangle(pos_x, pos_y, 2 * delta, 1)
        context.fill()

    # clear top rectangle for text:
    begin_bar = 0.5 - bar_height / 2
    end_bar = 0.5 + bar_height / 2
    context.set_operator(cairo.OPERATOR_SOURCE)
    context.set_source_rgba(0, 0, 0, 0)
    context.rectangle(0, 0, 1, begin_bar)
    context.fill()
    context.rectangle(0, end_bar, 1, end_bar)
    context.fill()

    # turn on antialiasing again for text:
    context.set_antialias(cairo.ANTIALIAS_SUBPIXEL)

    # tirn back on alpha blending:
    context.set_operator(cairo.OPERATOR_OVER)

    # draw text
    context.set_source_rgba(*color)
    context.select_font_face(font_name,
                             cairo.FONT_SLANT_NORMAL,
                             cairo.FONT_WEIGHT_NORMAL)
    context.set_font_size(font_size)

    text_height = context.text_extents('X')[3]

    start_text = number_format.format(start)
    context.move_to(0.01, end_bar + text_height / 2 + text_height)
    context.show_text(start_text)

    ext = context.text_extents(f"{title}")
    utw = ext[2]
    context.move_to(0.5 - utw / 2, begin_bar - text_height / 2)
    context.show_text(f"{title}")

    end_text = number_format.format(end)
    ext = context.text_extents(end_text)
    utw = ext[2]
    context.move_to(0.99 - utw, end_bar + text_height / 2 + text_height)
    context.show_text(end_text)

    # We remember where does the figure start and end vertically:
    vert_start = begin_bar - 3 * text_height / 2
    vert_end = end_bar + text_height / 2 + text_height

    surface_array = get_array_for_cairo_surface(surface)

    # Crop final image:
    height = surface_array.shape[0]
    crop_top = int((vert_start - text_height / 2) * height)
    crop_bottom = int((vert_end + text_height / 2) * height)
    surface_array = surface_array[crop_top - 10:crop_bottom + 10, ...]

    # Move to backend:
    surface_array = Backend.to_backend(surface_array, force_copy=True)

    return surface_array



import os
import random
import tempfile
from os.path import join

from arbol import aprint, asection
from dask.array.image import imread
from skimage.color import gray2rgba
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.video.overlay import add_overlays_image_sequence


def demo_overlay_numpy():
    with NumpyBackend():
        demo_overlay(display=True)


def demo_overlay_cupy():
    try:
        with CupyBackend():
            demo_overlay(n=64, display=True)
            return True

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def demo_overlay(n=16, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        image = gray2rgba(camera())

        # generate reference 'ground truth' timelapse
        images = (Backend.to_backend(image.copy()) for _ in range(n))

        # modify each image:
        images = (sp.ndimage.shift(image, shift=(random.uniform(-1, 2), random.uniform(-2, 1), 0)) for image in images)

        # Convert back images to 8 bit:
        images = list(image.astype(xp.uint8) for image in images)

    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        with asection("Save image sequences..."):
            # Create two subfolders for the two image sequences:
            input_folder = join(tmpdir, 'input_images')
            os.makedirs(input_folder)

            # Save images from first sequence:
            for i, image_u in enumerate(images):
                from PIL import Image
                im = Image.fromarray(Backend.to_numpy(image_u))
                im.save(join(input_folder, f"frame_{i:05}.png"))

        with asection("Overlay..."):
            # Create output folder:
            output_folder = join(tmpdir, 'overlay')
            os.makedirs(output_folder)

            # Perform overlay:

            # def add_overlays_image_sequence(input_path: str,
            #                                 output_path: str = None,
            #                                 scale_bar: bool = True,
            #                                 scale_bar_length_in_unit: float = 1,
            #                                 scale_bar_pixel_scale: float = 1,
            #                                 scale_bar_bar_height: int = 4,
            #                                 scale_bar_translation: Union[str, Sequence[Tuple[Union[int, float], ...]]] = None,
            #                                 scale_bar_unit: str = 'μm',
            #                                 time_stamp: bool = True,
            #                                 time_stamp_start_time: float = 0,
            #                                 time_stamp_time_interval: float = 1,
            #                                 time_stamp_translation: Union[str, Sequence[Tuple[Union[int, float], ...]]] = None,
            #                                 time_stamp_unit: str = 's',
            #                                 margin: float = 1,
            #                                 color: Tuple[float, float, float, float] = None,
            #                                 number_format: str = '{:.1f}',
            #                                 font_name: str = "Helvetica",
            #                                 font_size: float = 32,
            #                                 mode: str = 'max',
            #                                 overwrite: bool = False,
            #                                 workers: int = -1,
            #                                 workersbackend: str = 'threading'):

            add_overlays_image_sequence(input_path=input_folder,
                                        output_path=output_folder,
                                        scale_bar_length_in_unit=100,
                                        time_stamp_time_interval=20 / 60.0,
                                        margin=1)

        # load images into dask arrays:
        input_images = imread(os.path.join(input_folder, 'frame_*.png'))
        output_images = imread(os.path.join(output_folder, 'frame_*.png'))

        if display:
            from napari import Viewer, gui_qt
            with gui_qt():
                def _c(array):
                    return Backend.to_numpy(array)

                viewer = Viewer(ndisplay=2)
                # viewer.add_image(_c(images), name='images', rgb=True)
                viewer.add_image(_c(input_images), name='input_images', rgb=True)
                viewer.add_image(_c(output_images), name='output_images', rgb=True)

                viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_overlay_numpy():
        demo_overlay_cupy()



import click

from dexp.cli.dexp_main import _default_workers_backend
from dexp.video.overlay import add_overlays_image_sequence


@click.command()
@click.argument('input_path', type=str)
@click.option('--output_path', '-o', type=str, default=None, help='Output folder for overlayed frames.')
@click.option('--scalebar/--no-scalebar', '-sb/-nsb', default=True,
              help='True to insert scale bar.',
              show_default=True)
@click.option('--barlength', '-bl', type=float, default=1,
              help='Length of scale bar in the provided unit.',
              show_default=True)
@click.option('--barscale', '-bs', type=float, default=1,
              help='Conversion factor from pixels to units -- what is the side length of a pixel/voxel in units.',
              show_default=True)
@click.option('--barheight', '-bh', type=int, default=4,
              help='Height of th scale bar in pixels',
              show_default=True)
@click.option('--barpos', '-bp', type=str, default='bottom_right',
              help='Positions of the scale bar in pixels in natural order: (x, y).'
                   ' Can also be a string: bottom_left, bottom_right, top_left, top_right.',
              show_default=True)
@click.option('--barunit', '-bu', type=str, default='μm',
              help='Scale bar unit name.',
              show_default=True)
@click.option('--timestamp/--no-timestamp', '-ts/-nts', default=True,
              help='True to insert time stamp.',
              show_default=True)
@click.option('--timestart', '-ts', type=float, default=0,
              help='Start time for time stamp',
              show_default=True)
@click.option('--timeinterval', '-ti', type=float, default=1,
              help='Time interval inn units of time between consecutive images.',
              show_default=True)
@click.option('--timepos', '-tp', type=str, default='top_right',
              help='Positions of the time stamp in pixels in natural order: (x, y).'
                   ' Can also be a string: bottom_left, bottom_right, top_left, top_right.',
              show_default=True)
@click.option('--timeunit', '-tu', type=str, default='s',
              help='Time stamp time unit name.',
              show_default=True)
@click.option('--margin', '-mg', type=float, default=1,
              help='Margin around bar expressed in units relative to the text height',
              show_default=True)
@click.option('--color', '-c', type=str, default=None,
              help='Color of the bar and text as tuple of 4 values: (R, G, B, A)',
              show_default=True)
@click.option('--numberformat', '-nf', type=str, default='{:.1f}',
              help='Format string to represent the start and end values.',
              show_default=True)
@click.option('--fontname', '-fn', type=str, default='Helvetica',
              help='Font name.',
              show_default=True)
@click.option('--fontsize', '-fs', type=int, default=32,
              help='Font size in pixels.',
              show_default=True)
@click.option('--mode', '-md', type=str, default='max',
              help='Blending modes. Either one for all images, or one per image in the form of a sequence.'
                   ' Blending modes are: mean, add, satadd, max, alpha.',
              show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Force overwrite of output images.', show_default=True)
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, set to -1 for maximum number of workers', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--device', '-d', type=int, default=0, help='Sets the CUDA devices id, e.g. 0,1,2', show_default=True)  #
def overlay(input_path: str,
            output_path: str,
            scalebar,
            barlength,
            barscale,
            barheight,
            barpos,
            barunit,
            timestamp,
            timestart,
            timeinterval,
            timepos,
            timeunit,
            margin,
            color,
            numberformat,
            fontname,
            fontsize,
            mode,
            overwrite,
            workers,
            workersbackend,
            device):
    """ Adds scale bars and timestamps to videos.
    """

    # Default output path:
    if output_path is None:
        output_path = input_path + '_overlay'
    elif output_path.startswith('_'):
        output_path = input_path + output_path

    # Parse bar position:
    if ',' in barpos:
        barpos = tuple(float(v) for v in barpos.split(','))

    # Parse time position:
    if ',' in timepos:
        timepos = tuple(float(v) for v in timepos.split(','))

    # Parse color:
    if color is not None and ',' in color:
        color = tuple(float(v) for v in color.split(','))

    add_overlays_image_sequence(input_path=input_path,
                                output_path=output_path,
                                scale_bar=scalebar,
                                scale_bar_length_in_unit=barlength,
                                scale_bar_pixel_scale=barscale,
                                scale_bar_bar_height=barheight,
                                scale_bar_translation=barpos,
                                scale_bar_unit=barunit,
                                time_stamp=timestamp,
                                time_stamp_start_time=timestart,
                                time_stamp_time_interval=timeinterval,
                                time_stamp_translation=timepos,
                                time_stamp_unit=timeunit,
                                margin=margin,
                                color=color,
                                number_format=numberformat,
                                font_name=fontname,
                                font_size=fontsize,
                                mode=mode,
                                overwrite=overwrite,
                                workers=workers,
                                workersbackend=workersbackend,
                                device=device)



import click
from arbol.arbol import aprint

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group()
def cli():
    aprint("__________________________________________")
    aprint("  DEXP -- Data EXploration & Processing   ")
    aprint("  Video editing dexp_commands                  ")
    aprint("  Royer lab                               ")
    aprint("__________________________________________")
    aprint("")
    aprint("Note: videos are folders of images (typically PNGs), final step is to convert such image sequences into MP4s")
    # input("Press Enter to continue...")


from dexp.cli.video_commands.overlay import overlay
from dexp.cli.video_commands.blend import blend
from dexp.cli.video_commands.stack import stack
from dexp.cli.video_commands.resize import resize
from dexp.cli.video_commands.mp4 import mp4

cli.add_command(overlay)
cli.add_command(blend)
cli.add_command(stack)
cli.add_command(resize)
cli.add_command(mp4)



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.kernels.gaussian import gaussian_kernel_nd


def demo_kernels_numpy():
    with NumpyBackend():
        _demo_kernels()


def demo_kernels_cupy():
    try:
        with CupyBackend():
            _demo_kernels()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_kernels():
    gaussian_kernel = gaussian_kernel_nd(ndim=4, size=17, sigma=2)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(gaussian_kernel), name='gaussian_kernel')


demo_kernels_cupy()
demo_kernels_numpy()



from typing import Sequence, List

from arbol.arbol import aprint
from arbol.arbol import asection
from joblib import Parallel, delayed
from zarr.errors import ContainsArrayError, ContainsGroupError

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.multiview_lightsheet.fusion.mvsols import msols_fuse_1C2L
from dexp.processing.multiview_lightsheet.fusion.simview import simview_fuse_2C2L
from dexp.processing.registration.model.model_io import model_list_from_file, model_list_to_file


def dataset_fuse(dataset,
                 output_path,
                 channels,
                 slicing,
                 store,
                 compression,
                 compression_level,
                 overwrite,
                 microscope,
                 equalise,
                 equalise_mode,
                 zero_level,
                 clip_too_high,
                 fusion,
                 fusion_bias_strength_i,
                 fusion_bias_strength_d,
                 dehaze_size,
                 dark_denoise_threshold,
                 z_pad_apodise,
                 loadreg,
                 warpreg_num_iterations,
                 min_confidence,
                 max_change,
                 registration_edge_filter,
                 huge_dataset,
                 workers,
                 workersbackend,
                 devices,
                 check,
                 stop_at_exception=True):

    views = tuple(dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for channel in channels)

    with asection(f"views:"):
        for view, channel in zip(views, channels):
            aprint(f"View: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        views = tuple(view[slicing] for view in views)

    # shape and dtype of views to fuse:
    shape = views[0].shape
    dtype = views[0].dtype
    nb_timepoints = shape[0]

    # We allocate last minute once we know the shape...
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(output_path, mode, store)

    # load registration models:
    with NumpyBackend():
        model_list_filename = "registration_models.txt"
        if loadreg:
            aprint(f"Loading registration shifts from existing file! ({model_list_filename})")
            models = model_list_from_file(model_list_filename)
        else:
            models = [None, ] * nb_timepoints

    # hold equalisation ratios:
    equalisation_ratios_reference: List[Sequence[float]] = [[]]
    if microscope == 'simview':
        equalisation_ratios_reference[0] = [None, None, None]
    elif microscope == 'mvsols':
        equalisation_ratios_reference[0] = [None, ]

    def process(tp, device, workers):
        try:

            with asection(f"Loading channels {channels} for time point {tp}/{nb_timepoints}"):
                views_tp = tuple(view[tp].compute() for view in views)

            with BestBackend(device, exclusive=True, enable_unified_memory=True):

                model = models[tp]

                # If we don't have a model for that timepoint we load one from a previous timepoint
                if model is None and tp >= workers:
                    aprint(f"we don't have a registration model for timepoint {tp}/{nb_timepoints} so we load one from a previous timepoint: {tp - workers}")
                    model = models[tp - workers]

                if microscope == 'simview':
                    tp_array, model, new_equalisation_ratios = simview_fuse_2C2L(*views_tp,
                                                                                 registration_force_model=loadreg,
                                                                                 registration_model=model,
                                                                                 registration_min_confidence=min_confidence,
                                                                                 registration_max_change=max_change,
                                                                                 registration_edge_filter=registration_edge_filter,
                                                                                 equalise=equalise,
                                                                                 equalisation_ratios=equalisation_ratios_reference[0],
                                                                                 zero_level=zero_level,
                                                                                 clip_too_high=clip_too_high,
                                                                                 fusion=fusion,
                                                                                 fusion_bias_exponent=2,
                                                                                 fusion_bias_strength_i=fusion_bias_strength_i,
                                                                                 fusion_bias_strength_d=fusion_bias_strength_d,
                                                                                 dehaze_size=dehaze_size,
                                                                                 dark_denoise_threshold=dark_denoise_threshold,
                                                                                 huge_dataset_mode=huge_dataset)
                elif microscope == 'mvsols':
                    metadata = dataset.get_metadata()
                    angle = metadata['angle']
                    channel = metadata['channel']
                    dz = metadata['dz']
                    res = metadata['res']
                    illumination_correction_sigma = metadata['ic_sigma'] if 'ic_sigma' in metadata else None

                    tp_array, model, new_equalisation_ratios = msols_fuse_1C2L(*views_tp,
                                                                               z_pad=z_pad_apodise[0],
                                                                               z_apodise=z_pad_apodise[1],
                                                                               registration_num_iterations=warpreg_num_iterations,
                                                                               registration_force_model=loadreg,
                                                                               registration_model=model,
                                                                               registration_min_confidence=min_confidence,
                                                                               registration_max_change=max_change,
                                                                               registration_edge_filter=registration_edge_filter,
                                                                               equalise=equalise,
                                                                               equalisation_ratios=equalisation_ratios_reference[0],
                                                                               zero_level=zero_level,
                                                                               clip_too_high=clip_too_high,
                                                                               fusion=fusion,
                                                                               fusion_bias_exponent=2,
                                                                               fusion_bias_strength_x=fusion_bias_strength_i,
                                                                               dehaze_size=dehaze_size,
                                                                               dark_denoise_threshold=dark_denoise_threshold,
                                                                               angle=angle,
                                                                               dx=res,
                                                                               dz=dz,
                                                                               illumination_correction_sigma=illumination_correction_sigma,
                                                                               huge_dataset_mode=huge_dataset)

                with asection(f"Moving array from backend to numpy."):
                    tp_array = Backend.to_numpy(tp_array, dtype=dtype, force_copy=False)

                models[tp] = model.to_numpy()

                aprint(f"Last equalisation ratios: {new_equalisation_ratios}")
                if equalise_mode == 'first' and tp == 0:
                    aprint(f"Equalisation mode: 'first' -> saving equalisation ratios: {new_equalisation_ratios} for subsequent time points")
                    equalisation_ratios_reference[0] = list(float(Backend.to_numpy(ratio)) for ratio in new_equalisation_ratios)
                elif equalise_mode == 'all':
                    aprint(f"Equalisation mode: 'all' -> recomputing equalisation ratios for each time point.")
                    # No need to save, we need to recompute the ratios for each time point.
                    pass

            if 'fused' not in dest_dataset.channels():
                try:
                    dest_dataset.add_channel('fused',
                                             shape=(nb_timepoints,) + tp_array.shape,
                                             dtype=dtype,
                                             codec=compression,
                                             clevel=compression_level)
                except (ContainsArrayError, ContainsGroupError):
                    aprint(f"Other thread/process created channel before... ")

            with asection(f"Saving fused stack for time point {tp}, shape:{tp_array.shape}, dtype:{tp_array.dtype}"):
                dest_dataset.write_stack(channel='fused',
                                         time_point=tp,
                                         stack_array=tp_array)

            aprint(f"Done processing time point: {tp}/{nb_timepoints} .")

        except Exception as error:
            aprint(error)
            aprint(f"Error occurred while processing time point {tp} !")
            import traceback
            traceback.print_exc()

            if stop_at_exception:
                raise error

    if workers == -1:
        workers = len(devices)
    aprint(f"Number of workers: {workers}")

    if workers > 1:
        Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp, devices[tp % len(devices)], workers) for tp in range(0, nb_timepoints))
    else:
        for tp in range(0, nb_timepoints):
            process(tp, devices[0], workers)

    if not loadreg:
        model_list_to_file(model_list_filename, models)

    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.binary_blobs import binary_blobs
from dexp.utils.timeit import timeit


def demo_blobs_numpy():
    with NumpyBackend():
        demo_blobs()


def demo_blobs_cupy():
    try:
        with CupyBackend():
            demo_blobs()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_blobs(length_xy=320):
    with timeit("generate data"):
        image_blobs = binary_blobs(length=length_xy, n_dim=3, blob_size_fraction=0.07, volume_fraction=0.1).astype('f4')

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_blobs), name='image_gt')


if __name__ == "__main__":
    demo_blobs_cupy()
    demo_blobs_numpy()



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_clevel, _default_codec, _default_store
from dexp.cli.parsing import _get_output_path, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.isonet import dataset_isonet


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')  #
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--codec', '-z', default=_default_codec, help='compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ', show_default=True)  #
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--context', '-c', default='default', help="IsoNet context name", show_default=True)  # , help='dataset slice'
@click.option('--mode', '-m', default='pta', help="mode: 'pta' -> prepare, train and apply, 'a' just apply  ", show_default=True)  # , help='dataset slice'
@click.option('--max_epochs', '-e', type=int, default='50', help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def isonet(input_paths,
           output_path,
           slicing,
           store,
           codec,
           clevel,
           overwrite,
           context,
           mode,
           max_epochs,
           check):
    """ Recovers isotropic resolution using the ISONET approach (Weigert et al.)
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, '_isonet')
    slicing = _parse_slicing(slicing)

    with asection(f"Applying Isonet to: {input_paths}, saving result to: {output_path}, slicing: {slicing} "):
        dataset_isonet(input_dataset,
                       output_path,
                       slicing=slicing,
                       store=store,
                       compression=codec,
                       compression_level=clevel,
                       overwrite=overwrite,
                       context=context,
                       mode=mode,
                       max_epochs=max_epochs,
                       check=check
                       )

        input_dataset.close()
        aprint("Done!")



import numpy
from arbol import asection

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.projection import project_image
from dexp.processing.color.projection_legend import depth_color_scale_legend
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_projection_numpy():
    # TODO: projection behaves differently in numpy ?!? back and front views look the same ?!!
    with NumpyBackend():
        demo_projection(length_xy=150)


def demo_projection_cupy():
    try:
        with CupyBackend():
            demo_projection(length_xy=320)
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_projection(length_xy=120, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        _, _, image = generate_nuclei_background_data(add_noise=False,
                                                      length_xy=length_xy,
                                                      length_z_factor=1,
                                                      background_stength=0.001,
                                                      sphere=True,
                                                      radius=0.7,
                                                      zoom=2,
                                                      dtype=numpy.uint16)

        # we place an indicator to know where is the depth 0, will appear at the bottom right of the image:
        image[0, 2 * length_xy - 20:2 * length_xy, 2 * length_xy - 20:2 * length_xy] = 300

        # we place an indicator to know where is the depth max, will appear at the top right of the image:
        image[image.shape[0] - 1, 0:20, 2 * length_xy - 20:2 * length_xy] = 300

        # we place an indicator on top of the sample:
        image[image.shape[0] - 5:image.shape[0] - 1, :, length_xy - 10:length_xy + 10] = 300

    with asection("max_projection"):
        max_projection = project_image(image,
                                       mode='max')

    with asection("max_projection_att"):
        max_projection_att = project_image(image,
                                           mode='max',
                                           attenuation=0.05)

    with asection("max_color_projection"):
        max_color_projection = project_image(image,
                                             mode='maxcolor',
                                             attenuation=0.05,
                                             cmap='rainbow',
                                             dlim=(0.3, 0.7),
                                             legend_size=0.2
                                             )

    with asection("max_color_projection_bottom"):
        max_color_projection_bottom = project_image(image,
                                                    mode='maxcolor',
                                                    attenuation=0.05,
                                                    cmap='rainbow',
                                                    dlim=(0.3, 0.7),
                                                    dir=+1,
                                                    legend_size=0.2
                                                    )

    with asection("color_max_projection"):
        color_max_projection = project_image(image,
                                             mode='colormax',
                                             attenuation=0.1,
                                             cmap='rainbow',
                                             dlim=(0.3, 0.7),
                                             legend_size=0.2
                                             )

    with asection("color_max_projection_bottom"):
        color_max_projection_bottom = project_image(image,
                                                    mode='colormax',
                                                    attenuation=0.1,
                                                    cmap='rainbow',
                                                    dlim=(0.3, 0.7),
                                                    dir=+1,
                                                    legend_size=0.2
                                                    )

    with asection("color_max_projection_transparent"):
        color_max_projection_transparent = project_image(image,
                                                         mode='colormax',
                                                         attenuation=0.1,
                                                         cmap='rainbow',
                                                         dlim=(0.3, 0.7),
                                                         legend_size=0.2,
                                                         transparency=True,
                                                         )

    color_legend = depth_color_scale_legend(cmap='rainbow',
                                            start=0,
                                            end=100,
                                            title='color-coded depth (μm)',
                                            size=1)

    # image_shifted = sp.ndimage.shift(image, shift=(50, 70, -23), order=1, mode='nearest')

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image')
            viewer.add_image(_c(color_legend), name='color_legend', rgb=True)
            viewer.add_image(_c(max_projection), name='max_projection', rgb=True)
            viewer.add_image(_c(max_projection_att), name='max_projection_att', rgb=True)
            viewer.add_image(_c(color_max_projection), name='color_max_projection', rgb=True)
            viewer.add_image(_c(color_max_projection_bottom), name='color_max_projection_bottom', rgb=True)
            viewer.add_image(_c(max_color_projection), name='max_color_projection', rgb=True)
            viewer.add_image(_c(max_color_projection_bottom), name='max_color_projection_bottom', rgb=True)
            viewer.add_image(_c(color_max_projection_transparent), name='color_max_projection_transparent', rgb=True)
            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_projection_cupy():
        demo_projection_numpy()



import numpy
from skimage.data import camera
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.filters.kernels.gaussian import gaussian_kernel_nd


def test_lr_deconvolution_numpy():
    with NumpyBackend():
        _test_lr_deconvolution()


def test_lr_deconvolution_cupy():
    try:
        with CupyBackend():
            _test_lr_deconvolution()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_lr_deconvolution():
    xp = Backend.get_xp_module()

    image = camera().astype(numpy.float32) / 255
    noisy = random_noise(image, mode="gaussian", var=0.005, seed=0, clip=False)
    noisy = random_noise(noisy, mode="s&p", amount=0.03, seed=0, clip=False)

    psf = gaussian_kernel_nd(size=9, ndim=2, sigma=2, dtype=numpy.float32)

    image = Backend.to_backend(image)
    noisy = Backend.to_backend(noisy)
    psf = Backend.to_backend(psf)

    blurry = fft_convolve(image, psf)

    deconvolved = lucy_richardson_deconvolution(blurry, psf, padding=16)

    error = xp.linalg.norm(deconvolved - image, ord=1) / image.size

    print(f"Error = {error}")

    assert error < 0.001



import numpy
from arbol import aprint, asection
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.translation_nd import register_translation_nd


def demo_register_translation_2d_numpy():
    with NumpyBackend():
        _register_translation_2d()


def demo_register_translation_2d_cupy():
    try:
        with CupyBackend():
            _register_translation_2d()
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")


def _register_translation_2d(shift=(13, -5), display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate dataset"):
        image = camera().astype(numpy.float32) / 255
        image = Backend.to_backend(image)
        image = image[0:511, 0:501]

    with asection("shift"):
        shifted = sp.ndimage.shift(image, shift=shift)
        aprint(f"shift applied: {shift}")

    with asection("register_translation_2d"):
        model = register_translation_nd(image, shifted)
        aprint(f"model: {model}")

    with asection("shift back"):
        _, unshifted = model.apply_pair(image, shifted)
        image1_reg, image2_reg = model.apply_pair(image, shifted, pad=False)
        image1_reg_pad, image2_reg_pad = model.apply_pair(image, shifted, pad=True)

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', visible=False)
            viewer.add_image(_c(shifted), name='shifted', colormap='bop blue', blending='additive', visible=False)
            viewer.add_image(_c(unshifted), name='unshifted', colormap='bop blue', blending='additive', visible=False)
            viewer.add_image(_c(image1_reg), name='image1_reg', colormap='bop orange', blending='additive', visible=False)
            viewer.add_image(_c(image2_reg), name='image2_reg', colormap='bop blue', blending='additive', visible=False)
            viewer.add_image(_c(image1_reg_pad), name='image1_reg_pad', colormap='bop orange', blending='additive')
            viewer.add_image(_c(image2_reg_pad), name='image2_reg_pad', colormap='bop blue', blending='additive')

    return image, shifted, unshifted, model


if __name__ == "__main__":
    demo_register_translation_2d_cupy()
    demo_register_translation_2d_numpy()



import numpy
from scipy.ndimage import zoom

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def warp(image,
         vector_field,
         vector_field_upsampling: int = 2,
         vector_field_upsampling_order: int = 1,
         mode: str = 'border',
         image_to_backend: bool = False,
         internal_dtype=None):
    """
    Applies a warp transform (piece wise linear or constant) to an image based on a vector field.
    Only implemented for 1d, 2d, and 3d images.

    Parameters
    ----------
    image : image to warp

    vector_field : vector field to warp inoput image with. The vector field is an array of dimension n+1 where n is the dimension of the input image.
    The first n dimensions can be of arbirary lengths, and the last vector is the warp vector for each image region that the first

    vector_field_upsampling : upsampling factor for teh vector field (best use a power of two)

    vector_field_upsampling_order : upsampling order: 0-> nearest, 1->linear, 2->quadratic, ... (uses scipy zoom)

    mode : How to handle warping that reaches outside of the image bounds, can be: 'clamp', 'border', 'wrap', 'mirror'

    image_to_backend : By default one can directly copy a numpy array to texture memory,
    if needed, this option let's one first more the data to a cupy array before moving to texture memory. Not recommended.

    internal_dtype : internal dtype. Right now the dtype must be float32 because of CUDa texture dtype limitations.

    Returns
    -------
    Warped image

    """

    if not (image.ndim + 1 == vector_field.ndim or (image.ndim == 1 and vector_field.ndim == 1)):
        raise ValueError("Vector field must have one additional dimension")

    if internal_dtype is None:
        internal_dtype = numpy.float32

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype

    if vector_field_upsampling != 1:
        # Note: unfortunately numpy does support float16 zooming, and cupy does not support high-order zooming...
        vector_field = Backend.to_numpy(vector_field, dtype=numpy.float32)
        if image.ndim > 1:
            vector_field = zoom(vector_field, zoom=(vector_field_upsampling,) * image.ndim + (1,), order=vector_field_upsampling_order)
        else:
            vector_field = zoom(vector_field, zoom=(vector_field_upsampling,), order=vector_field_upsampling_order)

    # we can actually directly copy from numpy to texture mem!
    if image_to_backend:
        image = Backend.to_backend(image, dtype=internal_dtype)
    image = image.astype(dtype=internal_dtype, copy=False)
    vector_field = Backend.to_backend(vector_field, dtype=internal_dtype)

    from dexp.processing.backends.cupy_backend import CupyBackend
    if type(Backend.current()) is NumpyBackend:
        raise NotImplementedError("Warping not yet implemented for the Numpy backend.")

    elif type(Backend.current()) is CupyBackend:

        params = (image, vector_field, mode)
        if image.ndim == 1:
            from dexp.processing.interpolation._cupy.warp_1d import _warp_1d_cupy
            result = _warp_1d_cupy(*params)
        elif image.ndim == 2:
            from dexp.processing.interpolation._cupy.warp_2d import _warp_2d_cupy
            result = _warp_2d_cupy(*params)
        elif image.ndim == 3:
            from dexp.processing.interpolation._cupy.warp_3d import _warp_3d_cupy
            result = _warp_3d_cupy(*params)
        else:
            raise NotImplementedError("Warping for ndim>3 not implemented.")

    result = result.astype(original_dtype, copy=False)

    return result



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.kernels.butterworth import butterworth_kernel


def demo_butterworth_numpy():
    with NumpyBackend():
        _demo_butterworth()


def demo_butterworth_cupy():
    try:
        with CupyBackend():
            _demo_butterworth()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_butterworth():
    xp = Backend.get_xp_module()

    b = butterworth_kernel(shape=(31, 31),
                           cutoffs=0.75,
                           cutoffs_in_freq_units=False,
                           epsilon=1,
                           order=7)

    b_f = xp.log1p(xp.absolute(xp.fft.fftshift(xp.fft.fftn(b))))

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(b), name='b', colormap='viridis')
        viewer.add_image(_c(b_f), name='b_f', colormap='viridis')
        viewer.grid.enabled = True
        viewer.grid.shape = (2, 2)


demo_butterworth_cupy()
demo_butterworth_numpy()



import numexpr

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend


def element_wise_affine(array, alpha, beta, sum_first=False, out=None):
    """
    Applies the affine function: alpha*x + beta to every value x of a given array. If sum_first is True, then alpha*(x + beta) is computed instead.

    Parameters
    ----------
    array : array to apply function to
    alpha : 'scale'
    beta : 'offset'
    sum_first : If True then alpha*(x + beta) is computed instead.
    out : if specified, element_wise_affine _might_ use the specified array as output for in-place operation.
          this is not garanteed, so you should always use the pattern: result = element_wise_affine(backend, a, array, alpha, beta, out=result).


    Returns
    -------
    Array: alpha*array + beta (or alpha*array + beta if sum_first is True)

    """

    array = Backend.to_backend(array)

    if type(Backend.current()) is NumpyBackend:
        if sum_first:
            return numexpr.evaluate("alpha*(array+beta)", casting='same_kind', out=out)
        else:
            return numexpr.evaluate("alpha*array+beta", casting='same_kind', out=out)

    elif type(Backend.current()) is CupyBackend:
        import cupy
        if sum_first:
            @cupy.fuse()
            def affine_function(_array, _alpha, _beta):
                return _alpha * (_array + _beta)

            return affine_function(array, alpha, beta)
        else:
            @cupy.fuse()
            def affine_function(_array, _alpha, _beta):
                return _alpha * _array + _beta

            return affine_function(array, alpha, beta)



from scipy.ndimage import gaussian_filter
from skimage.data import binary_blobs

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.restoration.aap_correction import axis_aligned_pattern_correction
from dexp.processing.restoration.test.test_aap_correction import add_patterned_noise


def demo_aap_correction_numpy():
    with NumpyBackend():
        demo_aap_correction()


def demo_aap_correction_cupy():
    try:
        with CupyBackend():
            demo_aap_correction()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def demo_aap_correction(length_xy=128, level=0.3):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image = binary_blobs(length=length_xy, seed=1, n_dim=3, volume_fraction=0.01)
    image = image.astype(xp.float32)
    image = gaussian_filter(image, sigma=4)
    image = add_patterned_noise(image, length_xy)

    corrected = axis_aligned_pattern_correction(image,
                                                in_place=False)

    import napari

    with napari.gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = napari.Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(corrected), name='corrected')
        viewer.grid.enabled = True


if __name__ == "__main__":
    demo_aap_correction_cupy()
    demo_aap_correction_numpy()



import numpy
import scipy
from arbol import asection, aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.warp_nd import register_warp_nd
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_register_warp_3d_numpy():
    with NumpyBackend():
        _register_warp_3d()


def demo_register_warp_3d_cupy():
    try:
        with CupyBackend():
            _register_warp_3d()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def _register_warp_3d(length_xy=256, warp_grid_size=3, reg_grid_size=6, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    _, _, image = generate_nuclei_background_data(add_noise=False,
                                                  length_xy=length_xy,
                                                  length_z_factor=1,
                                                  independent_haze=True,
                                                  sphere=True,
                                                  zoom=2,
                                                  dtype=numpy.float32)

    image = image[0:length_xy * 2 - 3, 0:length_xy * 2 - 5, 0:length_xy * 2 - 7]

    with asection("warp"):
        magnitude = 10
        vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3 + (3,))
        warped = warp(image, vector_field, vector_field_upsampling=8)
        aprint(f"vector field applied: {vector_field}")

    with asection("add noise"):
        image += xp.random.uniform(0, 40, size=image.shape)
        warped += xp.random.uniform(0, 40, size=warped.shape)

    with asection(f"register_warp_nd"):
        chunks = tuple(s // reg_grid_size for s in image.shape)
        margins = tuple(max(4, c // 3) for c in chunks)
        aprint(f"chunks={chunks}, margins={margins}")
        model = register_warp_nd(image, warped, chunks=chunks, margins=margins)
        model.clean()
        # print(f"vector field found: {vector_field}")

    with asection("unwarp"):
        _, unwarped = model.apply_pair(image, warped, vector_field_upsampling=4)

    vector_field = scipy.ndimage.zoom(vector_field, zoom=(2, 2, 2, 1), order=1)

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', rendering='attenuated_mip', attenuation=0.01)
            viewer.add_image(_c(warped), name='warped', colormap='bop blue', blending='additive', visible=False, rendering='attenuated_mip', attenuation=0.01)
            viewer.add_image(_c(unwarped), name='unwarped', colormap='bop purple', blending='additive', rendering='attenuated_mip', attenuation=0.01)
            viewer.add_vectors(_c(vector_field), name='gt vector_field')
            viewer.add_vectors(_c(model.vector_field), name='model vector_field')

    return image, warped, unwarped, model


if __name__ == "__main__":
    demo_register_warp_3d_cupy()
    # demo_register_warp_3d_numpy()



import os
from typing import Sequence

import numpy
from arbol.arbol import aprint, asection

from dexp.datasets.base_dataset import BaseDataset


def dataset_copy(dataset: BaseDataset,
                 dest_path: str,
                 channels: Sequence[str],
                 slicing,
                 store: str,
                 compression: str,
                 compression_level: int,
                 overwrite: bool,
                 zerolevel: int,
                 workers: int,
                 workersbackend: str,
                 check: bool):

    # Create destination dataset:
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(dest_path, mode, store)

    # Process each channel:
    for channel in dataset._selected_channels(channels):

        with asection(f"Copying channel {channel}:"):
            array = dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)

            if slicing is not None:
                array = array[slicing]

            shape = array.shape
            dtype = array.dtype
            chunks = ZDataset._default_chunks

            dest_dataset.add_channel(name=channel,
                                     shape=shape,
                                     dtype=dtype,
                                     chunks=chunks,
                                     codec=compression,
                                     clevel=compression_level)

            def process(tp):
                try:
                    aprint(f"Processing time point: {tp} ...")
                    tp_array = array[tp].compute()
                    if zerolevel != 0:
                        tp_array = numpy.array(tp_array)
                        tp_array = numpy.clip(tp_array, a_min=zerolevel, a_max=None, out=tp_array)
                        tp_array -= zerolevel
                    dest_dataset.write_stack(channel=channel,
                                             time_point=tp,
                                             stack_array=tp_array)
                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while copying time point {tp} !")

            from joblib import Parallel, delayed

            if workers == -1:
                workers = max(1, os.cpu_count() // abs(workers))

            aprint(f"Number of workers: {workers}")
            Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp) for tp in range(0, shape[0]))

    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



import numpy
from tifffile import imsave


def tiff_save(file, img, axes='ZYX', compress=0, **imsave_kwargs):
    """Save image in ImageJ-compatible TIFF format.

    Parameters
    ----------
    file : str
        File name
    img : numpy.ndarray
        Image
    imsave_kwargs : dict, optional
        Keyword arguments for :func:`tifffile.imsave`

    """

    # convert to imagej-compatible data type
    t = img.dtype
    if 'float' in t.name:
        t_new = numpy.float32
    elif 'uint' in t.name:
        t_new = numpy.uint16 if t.itemsize >= 2 else numpy.uint8
    elif 'int' in t.name:
        t_new = numpy.int16
    else:
        t_new = t
    img = img.astype(t_new, copy=False)
    if t != t_new:
        print("Converting data type from '%s' to ImageJ-compatible '%s'." % (t, numpy.dtype(t_new)))

    imsave_kwargs['imagej'] = True
    imsave(file, img, **imsave_kwargs, compress=compress, metadata={'axes': axes})  # ,



import numexpr

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend


def blend_images(array_a, array_b,
                 blend_map,
                 dtype=None):
    """
    Blends two arrays on the basis of a blend map. A value of 1 in the blend map signifies that the value of the first array is chosen,
    a value of 0 means that the value from teh second array is chosen instead. For values within ]0,1[ linear blending is performed.

    Parameters
    ----------
    array_a : first array to blend
    array_b : second array to blend
    blend_map : blend map with values within [0, 1]
    dtype   : dtype for returned array

    Returns
    -------
    Blended array.

    """

    if array_a.dtype != array_b.dtype:
        raise ValueError("Two arrays to blend must have same dtype!")

    if array_a.shape != array_b.shape:
        raise ValueError("Two arrays to blend must have same shape!")

    if dtype is None:
        dtype = array_a.dtype

    array_a = Backend.to_backend(array_a, dtype=dtype)
    array_b = Backend.to_backend(array_b, dtype=dtype)
    blend_map = Backend.to_backend(blend_map)

    backend = Backend.current()

    if type(backend) is NumpyBackend:
        a = array_a
        b = array_b
        m = blend_map
        return numexpr.evaluate("a*m+(1-m)*b")

    elif type(backend) is CupyBackend:
        import cupy
        @cupy.fuse()
        def blend_function(a, b, m):
            return a * m + (1 - m) * b

        return blend_function(array_a, array_b, blend_map)



import click
from arbol.arbol import aprint, asection

from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument('input_paths', nargs=-1)
def info(input_paths):
    """ Prints out extensive information about dataset.
    """

    input_dataset, input_paths = glob_datasets(input_paths)

    with asection(f"Information on dataset at: {input_paths}"):
        aprint(input_dataset.info())
        input_dataset.close()
        aprint("Done!")



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def demo_register_translation_nD_numpy():
    with NumpyBackend():
        register_translation_nD()


def demo_register_translation_nD_cupy():
    try:
        with CupyBackend():
            register_translation_nD()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def register_translation_nD(length_xy=320):
    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False,
                                                                                       shift=(1, 5, -13),
                                                                                       volume_fraction=0.5,
                                                                                       length_xy=length_xy,
                                                                                       length_z_factor=2)

    model = TranslationRegistrationModel(shift_vector=(-1, -5, 13))

    image1_reg, image2_reg = model.apply_pair(image1, image2, pad=False)
    image1_reg_pad, image2_reg_pad = model.apply_pair(image1, image2, pad=True)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_gt), name='image_gt', visible=False)
        viewer.add_image(_c(image1), name='image1', colormap='bop blue', blending='additive')
        viewer.add_image(_c(image2), name='image2', colormap='bop orange', blending='additive', visible=False)
        viewer.add_image(_c(image1_reg), name='image1_reg', colormap='bop blue', blending='additive', visible=False)
        viewer.add_image(_c(image2_reg), name='image2_reg', colormap='bop orange', blending='additive')
        viewer.add_image(_c(image1_reg_pad), name='image1_reg_pad', colormap='bop blue', blending='additive', visible=False)
        viewer.add_image(_c(image2_reg_pad), name='image2_reg_pad', colormap='bop orange', blending='additive')


if __name__ == "__main__":
    demo_register_translation_nD_cupy()
    demo_register_translation_nD_numpy()



from typing import Union, Tuple

from arbol import asection, aprint, section

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.warp_registration_model import WarpRegistrationModel
from dexp.processing.registration.warp_nd import register_warp_nd


@section("register_warp_multiscale_nd")
def register_warp_multiscale_nd(image_a,
                                image_b,
                                num_iterations: int = 3,
                                confidence_threshold: float = 0.5,
                                max_residual_shift: float = None,
                                margin_ratios: Union[float, Tuple[float, ...]] = 0.2,
                                min_chunk: int = 32,
                                min_margin: int = 4,
                                save_memory: bool = False,
                                force_numpy: bool = False,
                                **kwargs) -> WarpRegistrationModel:
    """
    Registers image_b to image_a using a multi-scale warp approach.


    Parameters
    ----------
    image_a : First image to register
    image_b : Second image to register
    num_iterations : Number of iterations: each iteration subdivides chunks by a factor 2.
    confidence_threshold : confidence threshold for chunk registration
    max_residual_shift : max shift in pixels for all iterations except the first.
    margin_ratios : ratio that determines the margin size relative to the chunk size.
    min_chunk : minimal chunk size
    min_margin : minimum margin size
    save_memory : Moves data out of device memory as soon as possible during iterations to avoid memory overflows for very large images, incurs a largely negligible performance cost.
    force_numpy: Forces output model to be allocated with numpy arrays.
    all additional kwargs are passed to register_warp_nd

    Returns
    -------
    WarpRegistrationModel

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if image_a.shape != image_b.shape:
        raise ValueError("Image must have same shape!")

    ndim = image_a.ndim

    if type(margin_ratios) is not tuple:
        margin_ratios = (margin_ratios,) * ndim

    vector_field = None
    confidence = None

    with asection(f"Starting multi-scale registration with num_iterations={num_iterations}"):
        aprint(f"Confidence threshold: {confidence_threshold}, max residual shif: {max_residual_shift}")

        for i in range(num_iterations):

            # pre-apply transform from previous iterations:
            if vector_field is not None:
                with asection(f"Applying transform from previous iteration (save_memory={save_memory})"):
                    model = WarpRegistrationModel(vector_field, force_numpy=force_numpy)
                    _, image = model.apply_pair(image_a, image_b)
                    if save_memory:
                        image = Backend.to_numpy(image)
            else:
                image = image_b

            with asection(f"register_warp_nd {i}"):
                nb_div = 2 ** i
                # Note: the trick below is a 'ceil division' ceil(u/v) == -(-u//v)
                chunks = tuple(max(min_chunk, -(-s // nb_div)) for s in image_a.shape)
                margins = tuple(max(min_margin, int(c * r)) for c, r in zip(chunks, margin_ratios))
                aprint(f"register iteration: {i}, div= {nb_div}, chunks={chunks}, margins={margins}")
                model = register_warp_nd(image_a, image, chunks=chunks, margins=margins, force_numpy=force_numpy, **kwargs)
                aprint(f"median confidence: {model.median_confidence()}")
                aprint(f"median shift magnitude: {model.median_shift_magnitude(confidence_threshold=confidence_threshold)}")
                eff_max_shift = max_residual_shift if (i > 0 and max_residual_shift is not None) else None
                model.clean(mode='median', confidence_threshold=confidence_threshold,
                            max_shift=eff_max_shift)

                model_vector_field = Backend.to_backend(model.vector_field)
                model_confidence = Backend.to_backend(model.confidence)

            # if True:
            #     scale_factors = tuple(ims / mcs for ims, mcs in zip(image.shape, model_confidence.shape))
            #     aprint(f"scale_factors: {scale_factors}")
            #     model_confidence_n = Backend.to_numpy(model_confidence)
            #     import scipy
            #     _scaled_model_confidence_n = scipy.ndimage.zoom(model_confidence_n, zoom=scale_factors, order=0)
            #
            #     from napari import Viewer, gui_qt
            #     with gui_qt():
            #         def _c(array):
            #             return Backend.to_numpy(array)
            #
            #         viewer = Viewer()
            #         viewer.add_image(_c(image_a), name='image_a', colormap='bop orange', blending='additive')
            #         viewer.add_image(_c(image_b), name='image_b', colormap='bop blue', blending='additive', visible=False)
            #         viewer.add_image(_c(image), name='image', colormap='bop blue', blending='additive')
            #         viewer.add_image(_c(_scaled_model_confidence_n), name='_scaled_model_confidence_n', colormap='viridis', blending='additive', opacity=0.3)

            if vector_field is None:
                with asection(f"Initialise vector field and confidence arrays."):
                    splits = tuple(s // max(min_chunk, -(-s // (2 ** (num_iterations - 1)))) for s in image_a.shape)
                    aprint(f"final resolution = {splits}")
                    vector_field = xp.zeros(shape=splits + (ndim,), dtype=model_vector_field.dtype)
                    confidence = xp.zeros(shape=splits, dtype=model_confidence.dtype)

            if model.mean_confidence() > confidence_threshold // 2:
                with asection(f"Updating vector field and confidence array"):
                    scale_factors = tuple(s / ms for ms, s in zip(model_confidence.shape, confidence.shape))
                    aprint(f"Scaling obtained vector field and confidence arrays by a factor {scale_factors}")

                    scaled_vector_field = sp.ndimage.zoom(model_vector_field, zoom=scale_factors + (1,), order=1)
                    vector_field += scaled_vector_field

                    scaled_confidence = sp.ndimage.zoom(model_confidence, zoom=scale_factors, order=1)
                    confidence = xp.maximum(confidence, scaled_confidence)
            else:
                aprint(f"Scale ignored!")
                break

    model = WarpRegistrationModel(vector_field=vector_field,
                                  confidence=confidence,
                                  force_numpy=force_numpy)

    return model



from arbol import aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.synthetic_datasets.multiview_data import generate_fusion_test_data


def test_sequence_model_numpy():
    with NumpyBackend():
        _test_sequence_model()


def test_sequence_model_cupy():
    try:
        with CupyBackend():
            _test_sequence_model()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def _test_sequence_model(length_xy=128):
    xp = Backend.get_xp_module()

    model = SequenceRegistrationModel()

    assert len(model) == 0

    model_list = list(TranslationRegistrationModel(shift_vector=[1 * (1 + 0.2 * i), 5 + 0.5 * i, -13 - i], confidence=0.6) for i in range(0, 10))
    model = SequenceRegistrationModel(model_list)
    assert len(model) == 10

    image_gt, image_lowq, blend_a, blend_b, image1, image2 = generate_fusion_test_data(add_noise=False,
                                                                                       shift=(-1, -5, 13),
                                                                                       volume_fraction=0.5,
                                                                                       length_xy=length_xy,
                                                                                       length_z_factor=2,
                                                                                       amount_low=0,
                                                                                       zero_level=0)

    reg_images = list(model.apply(image2, index=index, pad=False) for index in range(0, 10))

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return Backend.to_numpy(array)
    #     viewer = Viewer()
    #     for i, reg_image in enumerate(reg_images):
    #         viewer.add_image(_c(reg_image), name=f'reg_image_{i}')

    average_errors = list(float(xp.mean(xp.absolute(reg_image - image1))) for reg_image in reg_images)
    aprint(average_errors)
    for u, e0 in enumerate(average_errors):
        for v, e1 in enumerate(average_errors):
            if u + 1 < v:
                assert e0 < e1

    reg_images_pad = list(model.apply(image2, index=index, pad=True) for index in range(0, 10))

    assert reg_images_pad[0].shape == model.padded_shape(image2.shape)
    assert model.padding() == ((0, 3), (0, 10), (22, 00))



from pytest import approx

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.demo.demo_translation_3d_proj import _register_translation_3d_proj
from dexp.processing.registration.translation_2d import register_translation_2d_skimage, register_translation_2d_dexp


def test_register_translation_3d_maxproj_numpy():
    with NumpyBackend():
        register_translation_3d_maxproj(register_translation_2d_skimage)
        register_translation_3d_maxproj(register_translation_2d_dexp)


def test_register_translation_3d_maxproj_cupy():
    try:
        with CupyBackend():
            register_translation_3d_maxproj(register_translation_2d_skimage)
            register_translation_3d_maxproj(register_translation_2d_dexp)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def register_translation_3d_maxproj(method, length_xy=128):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image, shifted, unshifted, model = _register_translation_3d_proj(length_xy=length_xy, method=method, display=False)
    shifts = Backend.to_numpy(model.shift_vector)
    assert shifts[0] == approx(-1, abs=0.5)
    assert shifts[1] == approx(-5, abs=0.5)
    assert shifts[2] == approx(13, abs=0.5)



import math
from typing import Tuple, Union

from dexp.processing.backends.backend import Backend
from dexp.processing.filters.butterworth_filter import butterworth_kernel
from dexp.processing.filters.kernels.wiener import wiener_kernel


def wiener_butterworth_kernel(kernel,
                              alpha: float = 1e-3,
                              beta: float = 1e-1,
                              cutoffs: Union[float, Tuple[float, ...], None] = None,
                              cutoffs_in_freq_units=False,
                              auto_cutoff_threshold=0.1,
                              order: int = 5,
                              dtype=None):
    """
    Computes the Wiener-Butterworth back projector according to Guo et al, bioRxiv 2019.

    Parameters
    ----------
    kernel : psf
    alpha : alpha
    beta : beta
    cutoffs : Butterworth cutoffs.
    cutoffs_in_freq_units : If True, the cutoffs are specified in frequency units. If False, the units are in normalised within [0,1]
    order : Butterworth order
    dtype : dtype for kernel

    Returns
    -------
    Wiener-Butterworth for given psf.

    """
    backend = Backend.current()
    xp = backend.get_xp_module()
    sp = backend.get_sp_module()

    if dtype is None:
        dtype = kernel.dtype

    wk_f = wiener_kernel(kernel,
                         alpha=alpha,
                         frequency_domain=True,
                         dtype=dtype)

    ##TODO: figure out cutoff from PSF ?

    if cutoffs is None:
        cutoffs_in_freq_units = False
        psf_f = xp.log1p(xp.absolute(xp.fft.fftshift(xp.fft.fftn(kernel))))

        psf_sumproj = []
        for i in range(psf_f.ndim):
            s = psf_f.shape[i]
            slicing = (s // 2,) * i + (slice(None),) + (s // 2,) * (psf_f.ndim - 1 - i)
            psf_sumproj.append(psf_f[slicing])

        psf_sumproj = tuple(p / p.max() for p in psf_sumproj)
        psf_sumproj = tuple(p[s // 2:] for s, p in zip(psf_f.shape, psf_sumproj))
        pass_band = tuple(p > auto_cutoff_threshold for p in psf_sumproj)
        cutoffs = tuple(float(xp.count_nonzero(b) / b.size) for b in pass_band)
        # cutoffs = (max(cutoffs),)*psf_f.ndim

    epsilon = math.sqrt((beta ** -2) - 1)

    bwk_f = butterworth_kernel(shape=kernel.shape,
                               cutoffs=cutoffs,
                               cutoffs_in_freq_units=cutoffs_in_freq_units,
                               epsilon=epsilon,
                               order=order,
                               frequency_domain=True,
                               dtype=dtype)

    # Weiner-Butterworth back projector
    wbwk_f = wk_f * bwk_f
    wbwk = xp.real(xp.fft.ifftn(wbwk_f))
    wbwk /= wbwk.sum()

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(xp.absolute(xp.fft.fftshift(array)))
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(wk_f), name='wk_f', colormap='viridis')
    #     viewer.add_image(_c(bwk_f), name='bwk_f', colormap='viridis')
    #     viewer.add_image(_c(wbwk_f), name='wbwk_f', colormap='viridis')
    #     viewer.grid_view(2, 2, 1)

    wbwk = wbwk.astype(dtype=dtype, copy=False)

    return wbwk

#
# ###
#
#    pfFFT = np.fft.fft2(pf)
#
#    # Wiener-Butterworth back projector.
#    #
#    # These values are from Guo et al.
#    alpha = 0.001
#    beta = 0.001
#    n = 8
#
#
#
#
#
#    # Wiener filter
#    bWiener = pfFFT/(np.abs(pfFFT) * np.abs(pfFFT) + alpha)
#
#    # Buttersworth filter
#    # kv = np.fft.fftfreq(pfFFT.shape[0])
#    # kx = np.zeros((kv.size, kv.size))
#    # for i in range(kv.size):
#    #     kx[i, :] = np.copy(kv)
#    # ky = np.transpose(kx)
#    # kk = np.sqrt(kx*kx + ky*ky)
#
#    # # This is the cut-off frequency.
#    # kc = 1.0/(0.5 * 2.355 * sigmaG)
#    # kkSqr = kk*kk/(kc*kc)
#    # eps = np.sqrt(1.0/(beta*beta) - 1)
#    # bBWorth = 1.0/np.sqrt(1.0 + eps * eps * np.power(kkSqr, n))
#
#    bw_kernel = butterworth_kernel(backend,
#                                   frequency_domain=True)
#
#    # Weiner-Butterworth back projector
#    pbFFT = bWiener * bBWorth
#
#    # back projector.
#    pb = np.real(np.fft.ifft2(pbFFT))
#
#    return [pf, pb]



import numpy
import pytest
from skimage.data import binary_blobs as binary_blobs_skimage

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.binary_blobs import binary_blobs as binary_blobs_dexp


def test_blobs_numpy():
    with NumpyBackend():
        _test_blobs()


def test_blobs_cupy():
    try:
        with CupyBackend():
            _test_blobs()
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")


def _test_blobs(length_xy=128):
    image_blobs = binary_blobs_dexp(length=length_xy, n_dim=3, blob_size_fraction=0.07, volume_fraction=0.1).astype('f4')
    image_blobs = Backend.to_numpy(image_blobs)

    image_blobs_skimage = binary_blobs_skimage(length=length_xy, n_dim=3, blob_size_fraction=0.07, volume_fraction=0.1).astype('f4')

    error = numpy.median(numpy.abs(image_blobs - image_blobs_skimage))
    print(error)
    assert error == pytest.approx(0)



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_store
from dexp.cli.parsing import _parse_channels, _get_output_path
from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='List of channels, all channels when ommited.')
@click.option('--rename', '-rc', default=None, help='You can rename channels: e.g. if channels are ‘channel1,anotherc’ then ‘gfp,rfp’ would rename the ‘channel1’ channel to ‘gfp’, and ‘anotherc’ to ‘rfp’ ')
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Forces overwrite of target', show_default=True)
def add(input_paths, output_path, channels, rename, store, overwrite):
    """ Adds the channels selected from INPUT_PATHS to the given output dataset (created if not existing).
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, '_add')
    channels = _parse_channels(input_dataset, channels)

    if rename is None:
        rename = input_dataset.channels()
    else:
        rename = rename.split(',')

    with asection(f"Adding channels: {channels} from: {input_paths} to {output_path}, with new names: {rename}"):
        input_dataset.add_channels_to(output_path,
                                      channels=channels,
                                      rename=rename,
                                      store=store,
                                      overwrite=overwrite
                                      )

        input_dataset.close()
        aprint("Done!")



import numpy


from dexp.processing.backends.backend import Backend


def get_array_for_cairo_surface(surface: 'ImageSurface'):
    """
    Returns an array given a ImageSurface from PyCairo.

    Parameters
    ----------
    surface : surface

    Returns
    -------
    RGBA numpy array of shape: (...,4)

    """

    from cairocffi import ImageSurface
    surface: ImageSurface


    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    width = surface.get_width()
    height = surface.get_height()

    # Get pycairo surface buffer:
    buffer = surface.get_data()

    # Reshape array to get an extra uint8 axis:
    surface_array = numpy.ndarray(shape=(height, width, 4), dtype=xp.uint8, buffer=buffer)

    # Move to backend:
    surface_array = Backend.to_backend(surface_array)

    # We have now: BGRA, we need to flip color axis because of endianness to ARGB:
    surface_array = xp.flip(surface_array, axis=surface_array.ndim - 1)

    # Convert ARGB to RGBA:
    surface_array = xp.roll(surface_array, shift=-1, axis=surface_array.ndim - 1)

    return surface_array



from typing import Callable

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.registration.translation_2d import register_translation_2d_dexp


def register_translation_proj_nd(image_a, image_b,
                                 register_translation_2d: Callable = register_translation_2d_dexp,
                                 drop_worse: bool = True,
                                 force_numpy: bool = False,
                                 internal_dtype=None,
                                 **kwargs):
    """
    Registers two nD (n=2 or 3) images using just a translation-only model.
    This method uses projections along 2 or 3 axis and then performs phase correlation.

    Parameters
    ----------
    image_a : First image to register
    image_b : Second image to register
    register_translation_2d : 2d registration method to use
    drop_worse: drops the worst 2D registrations before combining the projection registration vectors to a full nD registration vector.
    force_numpy : Forces output model to be allocated with numpy arrays.
    internal_dtype : Internal dtype for computation


    Returns
    -------
    Translation-only registration model

    """

    xp = Backend.get_xp_module()

    if image_a.ndim != image_b.ndim:
        raise ValueError("Images must have the same number of dimensions")

    image_a = Backend.to_backend(image_a)
    image_b = Backend.to_backend(image_b)

    if image_a.ndim == 2:
        image_a = _preprocess_image(image_a,
                                    in_place=False,
                                    dtype=internal_dtype)
        image_b = _preprocess_image(image_b,
                                    in_place=False,
                                    dtype=internal_dtype)

        shifts, confidence = register_translation_2d(image_a, image_b,
                                                     force_numpy=force_numpy,
                                                     internal_dtype=internal_dtype,
                                                     **kwargs).get_shift_and_confidence()

    elif image_a.ndim == 3:
        iap0 = _project_preprocess_image(image_a, axis=0, dtype=xp.float32)
        iap1 = _project_preprocess_image(image_a, axis=1, dtype=xp.float32)
        iap2 = _project_preprocess_image(image_a, axis=2, dtype=xp.float32)

        ibp0 = _project_preprocess_image(image_b, axis=0, dtype=xp.float32)
        ibp1 = _project_preprocess_image(image_b, axis=1, dtype=xp.float32)
        ibp2 = _project_preprocess_image(image_b, axis=2, dtype=xp.float32)

        shifts_p0, confidence_p0 = register_translation_2d(iap0, ibp0,
                                                           force_numpy=force_numpy,
                                                           internal_dtype=internal_dtype,
                                                           **kwargs).get_shift_and_confidence()

        shifts_p1, confidence_p1 = register_translation_2d(iap1, ibp1,
                                                           force_numpy=force_numpy,
                                                           internal_dtype=internal_dtype,
                                                           **kwargs).get_shift_and_confidence()

        shifts_p2, confidence_p2 = register_translation_2d(iap2, ibp2,
                                                           force_numpy=force_numpy,
                                                           internal_dtype=internal_dtype,
                                                           **kwargs).get_shift_and_confidence()

        # print(shifts_p0)
        # print(shifts_p1)
        # print(shifts_p2)

        if drop_worse:
            worse_index = xp.argmin(xp.asarray([confidence_p0, confidence_p1, confidence_p2]))

            if worse_index == 0:
                shifts = xp.asarray([0.5 * (shifts_p1[0] + shifts_p2[0]), shifts_p2[1], shifts_p1[1]])
                confidence = (confidence_p1 * confidence_p2) ** 0.5
            elif worse_index == 1:
                shifts = xp.asarray([shifts_p2[0], 0.5 * (shifts_p0[0] + shifts_p2[1]), shifts_p0[1]])
                confidence = (confidence_p0 * confidence_p2) ** 0.5
            elif worse_index == 2:
                shifts = xp.asarray([shifts_p1[0], shifts_p0[0], 0.5 * (shifts_p0[1] + shifts_p1[1])])
                confidence = (confidence_p0 * confidence_p1) ** 0.5

        else:
            shifts_p0 = xp.asarray([0, shifts_p0[0], shifts_p0[1]])
            shifts_p1 = xp.asarray([shifts_p1[0], 0, shifts_p1[1]])
            shifts_p2 = xp.asarray([shifts_p2[0], shifts_p2[1], 0])
            shifts = (shifts_p0 + shifts_p1 + shifts_p2) / 2
            confidence = (confidence_p0 * confidence_p1 * confidence_p2) ** 0.33

        # if confidence>0.1:
        #     print(f"shift={shifts}, confidence={confidence}")

        # from napari import Viewer, gui_qt
        # with gui_qt():
        #     def _c(array):
        #         return backend.to_numpy(array)
        #
        #     viewer = Viewer()
        #     viewer.add_image(_c(iap0), name='iap0')
        #     viewer.add_image(_c(ibp0), name='ibp0')
        #     viewer.add_image(_c(iap1), name='iap1')
        #     viewer.add_image(_c(ibp1), name='ibp1')
        #     viewer.add_image(_c(iap2), name='iap2')
        #     viewer.add_image(_c(ibp2), name='ibp2')
    else:
        raise ValueError(f'Unsupported number of dimensions ({image_a.ndim}) for registration.')

    model = TranslationRegistrationModel(shift_vector=shifts, confidence=confidence, force_numpy=force_numpy)

    return model


def _project_preprocess_image(image,
                              axis: int,
                              smoothing: float = 0,
                              quantile: int = None,
                              gamma: float = 1,
                              dtype=None):
    image_projected = _project_image(image, axis=axis)
    image_projected_processed = _preprocess_image(image_projected,
                                                  quantile=quantile,
                                                  dtype=dtype)

    return image_projected_processed


def _project_image(image, axis: int):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()
    image = Backend.to_backend(image)
    projection = xp.max(image, axis=axis) - xp.min(image, axis=axis)
    return projection


def _preprocess_image(image,
                      quantile: float = 0.01,
                      in_place: bool = True,
                      dtype=None):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    processed_image = Backend.to_backend(image, dtype=dtype, force_copy=not in_place)

    if quantile is None:
        min_value = processed_image.min()
        max_value = processed_image.max()
    else:
        min_value = xp.percentile(processed_image, q=100 * quantile)
        max_value = xp.percentile(processed_image, q=100 * (1 - quantile))

    alpha = max_value - min_value
    if alpha > 0:
        processed_image -= min_value
        processed_image /= alpha
        processed_image = xp.clip(processed_image, 0, 1, out=processed_image)

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image')
    #     viewer.add_image(_c(processed_image), name='processed_image')

    return processed_image



import ctypes
import os


def set_mkl_threads(num_threads=None):
    try:
        if num_threads is None:
            import multiprocessing
            num_threads = multiprocessing.cpu_count() // 2 - 1

        try:
            import mkl
            mkl.set_num_threads(num_threads)
            return 0
        except Exception as e:
            pass
            # import traceback
            # traceback.print_exc()
            # print(e)

        for name in ["libmkl_rt.so", "libmkl_rt.dylib", "mkl_Rt.dll"]:
            try:
                mkl_rt = ctypes.CDLL(name)
                mkl_rt.mkl_set_num_threads(ctypes.byref(ctypes.c_int(num_threads)))
                return 0
            except Exception as e:
                pass
                # import traceback
                # traceback.print_exc()
                # print(e)

        os.environ["OMP_NUM_THREADS"] = f"{num_threads}"  # export OMP_NUM_THREADS=4
        os.environ["OPENBLAS_NUM_THREADS"] = f"{num_threads}"  # export OPENBLAS_NUM_THREADS=4
        os.environ["MKL_NUM_THREADS"] = f"{num_threads}"  # export MKL_NUM_THREADS=6
        os.environ["VECLIB_MAXIMUM_THREADS"] = f"{num_threads}"  # export VECLIB_MAXIMUM_THREADS=4
        os.environ["NUMEXPR_NUM_THREADS"] = f"{num_threads}"  # export NUMEXPR_NUM_THREADS=6
    except Exception as e:
        pass
        # import traceback
        # traceback.print_exc()
        # print(e)

# set_mkl_threads(8)



import os
from os import listdir
from os.path import exists, isfile, join, isdir
from typing import Tuple, Sequence, Union

import imageio
from arbol.arbol import asection, aprint
from joblib import Parallel, delayed

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.color.blend import blend_color_images
from dexp.processing.color.insert import insert_color_image


def blend_color_image_sequences(input_paths: Sequence[str],
                                output_path: str = None,
                                modes: Union[str, Sequence[str]] = 'max',
                                alphas: Sequence[float] = None,
                                scales: Sequence[float] = None,
                                translations: Union[str, Sequence[Tuple[Union[int, float], ...]]] = None,
                                background_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
                                border_width: int = 1,
                                border_color: Tuple[float, float, float, float] = None,
                                border_over_image: bool = False,
                                overwrite: bool = False,
                                workers: int = -1,
                                workersbackend: str = 'threading',
                                device: int = 0):
    """
    Blends several RGB(A) image sequences together

    Parameters
    ----------
    input_paths : Paths to folders containing images in some (lexicographic) order, or to single images (that will be broadcasted). If of different lengths the result's length is the shortest of the input sequences.
    output_path : Path to save the blended images.
    modes : Blending modes for each input sequence.
    alphas : Alpha transparency applied to each input sequence.
    scales : Scaling ('zoom' in scipy parlance) applied to each input sequence.
    translations : Translation applied to each input sequence.
    background_color:  Background color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.
    border_width: Width of border added to insets.
    border_color: Border color.
    border_over_image: If True the border is not added but overlayed over the image, the image does not change size.
    overwrite : If True the output files are overwritten
    workers : Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)
    workersbackend : What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread)
    device: Device on  which to run the overlay computation (if a non-CPU device is available).
    """

    # ensure folder exists:
    os.makedirs(output_path, exist_ok=True)

    # collect all image files:
    image_sequences = []
    for input_path in input_paths:

        if isdir(input_path):
            # path is folder:
            pngfiles = [join(input_path, f) for f in listdir(input_path) if isfile(join(input_path, f)) and f.endswith('.png')]
            pngfiles.sort()

        elif isfile(input_path) and (input_path.endswith('png') or input_path.endswith('jpg')):
            # path is image file:
            pngfiles = [input_path, ]

        image_sequences.append(pngfiles)

    # Basic sanity check on the images sequence: we determine the shortest non-one length :
    min_length = min(len(image_sequence) for image_sequence in image_sequences if len(image_sequence) != 1)
    max_length = max(len(image_sequence) for image_sequence in image_sequences if len(image_sequence) != 1)
    if min_length != max_length:
        aprint(f"Not all image sequences have the same non-one length: min:{min_length}, max:{max_length}")

    # Now we broadcast and crop in time:
    _image_sequences = []
    for image_sequence in image_sequences:
        if len(image_sequence) == 1:
            image_sequence = [image_sequence[0], ] * min_length
        elif len(image_sequence) > min_length:
            image_sequence = image_sequence[0:min_length]
        _image_sequences.append(image_sequence)
    image_sequences = _image_sequences
    nb_timepoints = min_length

    with BestBackend(device, exclusive=True, enable_unified_memory=True):

        def _process(backend: Backend, tp: int):

            with asection(f'processing time point: {tp}'):

                # Output file:
                filename = f"frame_{tp:05}.png"
                filepath = join(output_path, filename)

                # Write file:
                if overwrite or not exists(filepath):

                    # We set the backend of this thread to be the same as its parent thread:
                    if backend is not None:
                        Backend.set(backend)

                    # collect all images that need to be blended:
                    image_paths = list(image_sequence[tp] for image_sequence in image_sequences)

                    # Load images:
                    images = list(imageio.imread(image_path) for image_path in image_paths)

                    if scales is None and translations is None:
                        # Blend images:
                        blended = blend_color_images(images=images,
                                                     alphas=alphas,
                                                     modes=modes,
                                                     background_color=background_color)
                    else:
                        xp = Backend.get_xp_module()
                        sp = Backend.get_sp_module()

                        # Prepare background image:
                        blended = xp.zeros(shape=images[0].shape,
                                           dtype=images[0].dtype)

                        # Fill with background color:
                        for channel in range(4):
                            blended[:, channel] = background_color[channel]

                        # Insert each image at a different location with different blending
                        for inset_image, mode, scale, alpha, trans in zip(images, modes, scales, alphas, translations):
                            blended = insert_color_image(image=blended,
                                                         inset_image=inset_image,
                                                         scale=scale,
                                                         translation=trans,
                                                         border_width=border_width,
                                                         border_color=border_color,
                                                         border_over_image=border_over_image,
                                                         mode=mode,
                                                         alpha=alpha,
                                                         background_color=(0, 0, 0, 0),
                                                         )

                    aprint(f"Writing file: {filename} in folder: {output_path}")
                    imageio.imwrite(filepath,
                                    Backend.to_numpy(blended),
                                    compress_level=1)
                else:
                    aprint(f"File: {filepath} already exists! use -w option to force overwrite...")

        with asection(f"Blending  {input_paths}, saving to {output_path}, for a total of {nb_timepoints} time points"):
            if workers == -1:
                workers = min(10, os.cpu_count() // 2)

            backend = Backend.current()

            if workers > 1:
                Parallel(n_jobs=workers, backend=workersbackend)(delayed(_process)(backend, tp) for tp in range(nb_timepoints))
            else:
                for tp in range(nb_timepoints):
                    _process(None, tp)

    aprint(f"Done!")



import gc
from typing import Tuple, Sequence

import numpy
from arbol.arbol import asection, aprint, section

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deskew.yang_deskew import yang_deskew
from dexp.processing.equalise.equalise_intensity import equalise_intensity
from dexp.processing.filters.butterworth_filter import butterworth_filter
from dexp.processing.fusion.dct_fusion import fuse_dct_nd
from dexp.processing.fusion.dft_fusion import fuse_dft_nd
from dexp.processing.fusion.tg_fusion import fuse_tg_nd
from dexp.processing.multiview_lightsheet.fusion.simview import fuse_illumination_views
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel
from dexp.processing.registration.translation_nd import register_translation_nd
from dexp.processing.registration.translation_nd_proj import register_translation_proj_nd
from dexp.processing.registration.warp_multiscale_nd import register_warp_multiscale_nd
from dexp.processing.restoration.clean_dark_regions import clean_dark_regions
from dexp.processing.restoration.dehazing import dehaze


@section("mvSOLS 1C2L fusion")
def msols_fuse_1C2L(C0L0, C0L1,
                    dz: float,
                    dx: float,
                    angle: float,
                    resampling_mode: str = 'yang',
                    equalise: bool = True,
                    equalisation_ratios: Sequence[float] = (None,),
                    zero_level: float = 120,
                    clip_too_high: int = 2048,
                    fusion='tg',
                    fusion_bias_exponent: int = 2,
                    fusion_bias_strength_x: float = 0.1,
                    z_pad: int = 0,
                    z_apodise: int = 0,
                    registration_num_iterations: int = 4,
                    registration_confidence_threshold: float = 0.3,
                    registration_max_residual_shift: int = 32,
                    registration_mode: str = 'projection',
                    registration_edge_filter: bool = False,
                    registration_force_model: bool = False,
                    registration_model: PairwiseRegistrationModel = None,
                    registration_min_confidence: float = 0.5,
                    registration_max_change: int = 16,
                    dehaze_before_fusion: bool = True,
                    dehaze_size: int = 65,
                    dehaze_correct_max_level: bool = True,
                    dark_denoise_threshold: int = 0,
                    dark_denoise_size: int = 9,
                    butterworth_filter_cutoff: float = 1,
                    illumination_correction_sigma: float = None,
                    huge_dataset_mode: bool = False,
                    internal_dtype=numpy.float16) -> Tuple:
    """

    Parameters
    ----------
    C0L0 : Image for Camera 0 lightsheet 0
    C0L1 : Image for Camera 0 lightsheet 1

    dz     : float, scanning step (stage or galvo scanning step, not the same as the distance between the slices)

    dx     : float, pixel size of the camera

    angle  : float, incident angle of the light sheet, angle between the light sheet and the optical axis

    mode : Resampling mode, can be 'yang' for Bin Yang's resampling ;-)

    equalise : Equalise intensity of views before fusion, or not.

    equalisation_ratios: If provided, these ratios are used instead of calculating equalisation values based on the images.
    There is only one equalisation ratio, therefore this is a singleton tuple: (ratio,)

    zero_level : Zero level: that's the minimal detector pixel value floor to substract,
    typically for sCMOS cameras the floor is at around 100 (this is to avoid negative values
    due to electronic noise!). Substracting a bit more than that is a good idea to clear out noise
    in the background --  hence the default of 120.

    clip_too_high : clips very high intensities, to avoid loss of precision when converting an internal format such as float16

    fusion : Fusion mode, can be 'tg', 'dct', 'dft'

    fusion_bias_exponent : Exponent for fusion bias

    fusion_bias_strength_x : Strength of fusion bias for fusing views along x axis (after resampling/deskewing) . Set to zero to deactivate

    z_pad : Padding length along Z (scanning direction) for input stacks, usefull in conjunction with z_apodise

    z_apodise : apodises along Z (direction) to suppress discontinuities (views cropping the sample) that disrupt fusion.

    registration_confidence_threshold : Confidence threshold used for each chunk during warp registration, value within [0, 1]: zero means low confidence, 1 max confidence.

    registration_max_residual_shift : After the first registration round of warp registration, shift vector with norms larger than this value are deemed low confidence.

    registration_mode : Registration mode, can be: 'projection' or 'full'.
    Projection mode is faster but might have occasionally  issues for certain samples. Full mode is slower and is only recomended as a last resort.

    registration_edge_filter : apply edge filter to help registration

    registration_force_model : Forces the use of the provided model (see below)

    registration_model : registration model to use the two camera views (C0Lx and C1Lx),
    if None, the two camera views are registered, and the registration model is returned.

    registration_min_confidence : Minimal confidence for registration parameters, if below that level the registration parameters for previous time points is used.

    registration_max_change : Maximal change in registration parameters, if above that level the registration parameters for previous time points is used.

    dehaze_before_fusion : Whether to dehaze the views before fusion or to dehaze the fully fused and registered final image.

    dehaze_size : After all fusion and registration, the final image is dehazed to remove
    large-scale background light caused by scattered illumination and out-of-focus light.
    This parameter controls the scale of the low-pass filter used.

    dehaze_correct_max_level : Should the dehazing correct the reduced local max intensity induced by removing the background?

    dark_denoise_threshold : After all fusion and registration, the final image is processed
    to remove any remaining noise in the dark background region (= hurts compression!).

    dark_denoise_size : Controls the scale over which pixels must be below the threshold to
    be categorised as 'dark background'.

    butterworth_filter_cutoff : At the very end, butterworth_filtering may be applied to
    smooth out spurious high frequencies in the final image. A value of 1 means no filtering,
    a value of e.g. 0.5 means cutting off the top 50% higher frequencies, and keeping all
    frequencies below without any change (that's the point of Butterworth filtering).
    WARNING: Butterworth filtering is currently very slow...

    illumination_correction_sigma: sigma in pixels for correcting the Gaussian profile of a cylindrical lightsheet.

    huge_dataset_mode: optimises memory allocation at the detriment of processing speed to tackle really huge datasets.

    internal_dtype : internal dtype


    Returns
    -------
    Fully registered, fused, dehazed 3D image

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if C0L0.dtype != C0L1.dtype:
        raise ValueError("The two views must have same dtype!")

    if C0L0.shape != C0L1.shape:
        raise ValueError("The two views must have same shapes!")

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = C0L0.dtype

    with asection(f"Moving C0L0 and C0L1 to backend storage and converting to {internal_dtype}..."):
        C0L0 = Backend.to_backend(C0L0, dtype=internal_dtype, force_copy=False)
        C0L1 = Backend.to_backend(C0L1, dtype=internal_dtype, force_copy=False)
        Backend.current().clear_memory_pool()

    if clip_too_high > 0:
        with asection(f"Clipping intensities above {clip_too_high} for C0L0 & C0L1"):
            C0L0 = xp.clip(C0L0, a_min=0, a_max=clip_too_high, out=C0L0)
            C0L1 = xp.clip(C0L1, a_min=0, a_max=clip_too_high, out=C0L1)
            Backend.current().clear_memory_pool()

    if z_pad > 0:
        with asection(f"Pad C0L0 and C0L1 along scanning direction:"):
            C0L0[0] = sp.ndimage.gaussian_filter(C0L0[0], sigma=4)
            C0L0[-1] = sp.ndimage.gaussian_filter(C0L0[-1], sigma=4)
            C0L0 = xp.pad(C0L0, pad_width=((z_pad, z_pad),) + ((0, 0),) * 2, mode='edge')
            Backend.current().clear_memory_pool()

            C0L1[0] = sp.ndimage.gaussian_filter(C0L1[0], sigma=4)
            C0L1[-1] = sp.ndimage.gaussian_filter(C0L1[-1], sigma=4)
            C0L1 = xp.pad(C0L1, pad_width=((z_pad, z_pad),) + ((0, 0),) * 2, mode='edge')
            Backend.current().clear_memory_pool()

    if z_apodise > 0:
        with asection(f"apodise C0L0 and C0L1 along scanning direction:"):
            depth = C0L0.shape[0]
            apodise_left = xp.linspace(0, 1, num=z_apodise, dtype=internal_dtype)
            apodise_left **= 3
            apodise_center = xp.ones(shape=(depth - 2 * z_apodise,), dtype=internal_dtype)
            apodise_right = xp.linspace(1, 0, num=z_apodise, dtype=internal_dtype) ** 0.5
            apodise_right **= 3
            apodise = xp.concatenate((apodise_left, apodise_center, apodise_right))
            apodise = apodise[:, xp.newaxis, xp.newaxis]
            apodise = apodise.astype(dtype=internal_dtype, copy=False)

            C0L0 *= apodise
            C0L1 *= apodise

            del apodise, apodise_left, apodise_right
            Backend.current().clear_memory_pool()

    # from napari import gui_qt, Viewer
    # with gui_qt():
    #     def _c(array):
    #         return Backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(C0L0), name='C0L0', contrast_limits=(0, 1000))
    #     viewer.add_image(_c(C0L1), name='C0L1', contrast_limits=(0, 1000))

    with asection(f"Resample C0L0 and C0L1"):
        C0L0 = resample_C0L0(C0L0, angle=angle, dx=dx, dz=dz, mode=resampling_mode)
        aprint(f"Shape and dtype of C0L0 after resampling: {C0L0.shape}, {C0L0.dtype}")
        C0L1 = resample_C0L1(C0L1, angle=angle, dx=dx, dz=dz, mode=resampling_mode)
        aprint(f"Shape and dtype of C0L1 after resampling: {C0L1.shape}, {C0L1.dtype}")
        Backend.current().clear_memory_pool()

    if dehaze_size > 0 and dehaze_before_fusion:
        with asection(f"Dehaze C0L0 & C0L1 ..."):
            C0L0 = dehaze(C0L0,
                          size=dehaze_size,
                          minimal_zero_level=zero_level,
                          correct_max_level=dehaze_correct_max_level)
            C0L1 = dehaze(C0L1,
                          size=dehaze_size,
                          minimal_zero_level=zero_level,
                          correct_max_level=dehaze_correct_max_level)
            Backend.current().clear_memory_pool()

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return Backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(C0L0), name='C0L0', colormap='bop blue', blending='additive')
    #     viewer.add_image(_c(C0L1), name='C0L1', colormap='bop orange', blending='additive')

    with asection(f"Register C0L0 and C0L1"):

        aprint(f"Provided registration model: {registration_model}, overall confidence: {0 if registration_model is None else registration_model.overall_confidence()}")

        if huge_dataset_mode:
            aprint(f"Huge dataset mode is on: moving data to CPU RAM.")
            C0L0 = Backend.to_numpy(C0L0)
            C0L1 = Backend.to_numpy(C0L1)

        if registration_force_model and registration_model is not None:
            model = registration_model
        else:
            aprint("No registration model enforced, running registration now")
            registration_method = register_translation_proj_nd if registration_mode == 'projection' else register_translation_nd
            new_model = register_warp_multiscale_nd(C0L0, C0L1,
                                                    num_iterations=registration_num_iterations,
                                                    confidence_threshold=registration_confidence_threshold,
                                                    max_residual_shift=registration_max_residual_shift,
                                                    edge_filter=registration_edge_filter,
                                                    registration_method=registration_method,
                                                    save_memory=huge_dataset_mode)

            Backend.current().clear_memory_pool()
            aprint(f"Computed registration model: {new_model}, overall confidence: {new_model.overall_confidence()}")

            if registration_model is None or (new_model.overall_confidence() >= registration_min_confidence or new_model.change_relative_to(registration_model) <= registration_max_change):
                model = new_model
            else:
                model = registration_model

        aprint(f"Applying registration model: {model}, overall confidence: {model.overall_confidence()}")
        C0L0, C0L1 = model.apply_pair(C0L0, C0L1)
        Backend.current().clear_memory_pool()

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return Backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(C0L0), name='C0L0', colormap='bop blue', blending='additive')
    #     viewer.add_image(_c(C0L1), name='C0L1', colormap='bop orange', blending='additive')

    if equalise:
        with asection(f"Equalise intensity of C0L0 relative to C0L1 ..."):
            C0L0, C0L1, ratio = equalise_intensity(C0L0, C0L1,
                                                   zero_level=0 if dehaze_before_fusion else zero_level,
                                                   correction_ratio=equalisation_ratios[0],
                                                   copy=False)
            equalisation_ratios = (ratio,)
            aprint(f"Equalisation ratio: {ratio}")
            Backend.current().clear_memory_pool()

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return Backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(C0L0), name='C0L0', colormap='bop blue', blending='additive')
    #     viewer.add_image(_c(C0L1), name='C0L1', colormap='bop orange', blending='additive')

    with asection(f"Fuse detection views C0lx and C1Lx..."):
        C1Lx = fuse_illumination_views(C0L0, C0L1,
                                       mode=fusion,
                                       bias_exponent=fusion_bias_exponent,
                                       bias_strength=fusion_bias_strength_x)
        Backend.current().clear_memory_pool()

    if dehaze_size > 0 and not dehaze_before_fusion:
        with asection(f"Dehaze CxLx ..."):
            C1Lx = dehaze(C1Lx,
                          size=dehaze_size,
                          minimal_zero_level=0,
                          correct_max_level=dehaze_correct_max_level)
            Backend.current().clear_memory_pool()

    if dark_denoise_threshold > 0:
        with asection(f"Denoise dark regions of CxLx..."):
            C1Lx = clean_dark_regions(C1Lx,
                                      size=dark_denoise_size,
                                      threshold=dark_denoise_threshold)
            Backend.current().clear_memory_pool()

    if 0 < butterworth_filter_cutoff < 1:
        with asection(f"Filter output using a Butterworth filter"):
            cutoffs = (butterworth_filter_cutoff,) * C1Lx.ndim
            C1Lx = butterworth_filter(C1Lx, shape=(31, 31, 31), cutoffs=cutoffs, cutoffs_in_freq_units=False)
            Backend.current().clear_memory_pool()

    if illumination_correction_sigma is not None:
        sigma = illumination_correction_sigma
        length = C1Lx.shape[1]
        correction = xp.linspace(-length // 2, length // 2, num=length)
        correction = xp.exp(-(correction ** 2) / (2 * sigma * sigma))
        correction = 1.0 / correction
        correction = correction.astype(dtype=internal_dtype)
        C1Lx *= correction[xp.newaxis, :, xp.newaxis]

    with asection(f"Convert back to original dtype..."):
        if original_dtype is numpy.uint16:
            C1Lx = xp.clip(C1Lx, 0, None, out=C1Lx)
        C1Lx = C1Lx.astype(dtype=original_dtype, copy=False)
        Backend.current().clear_memory_pool()

    gc.collect()
    Backend.current().clear_memory_pool()

    return C1Lx, model, equalisation_ratios


def fuse_illumination_views(CxL0, CxL1,
                            mode,
                            bias_exponent,
                            bias_strength,
                            smoothing: int = 12):
    if mode == 'tg':
        fused = fuse_tg_nd(CxL0, CxL1,
                           downscale=2,
                           tenengrad_smoothing=smoothing,
                           bias_axis=2,
                           bias_exponent=bias_exponent,
                           bias_strength=bias_strength)
    elif mode == 'dct':
        fused = fuse_dct_nd(CxL0, CxL1)
    elif mode == 'dft':
        fused = fuse_dft_nd(CxL0, CxL1)

    return fused


def resample_C0L0(image,
                  dx: float,
                  dz: float,
                  angle: float,
                  mode: str = 'yang',
                  num_split: int = 8):
    if mode == 'yang':
        return yang_deskew(image,
                           flip=True,
                           dx=dx,
                           dz=dz,
                           angle=angle,
                           num_split=num_split)
    else:
        raise ValueError("Invalid deskew mode")


def resample_C0L1(image,
                  dx: float,
                  dz: float,
                  angle: float,
                  mode: str = 'yang',
                  num_split: int = 8):
    if mode == 'yang':
        return yang_deskew(image,
                           flip=False,
                           dx=dx,
                           dz=dz,
                           angle=angle,
                           num_split=num_split)
    else:
        raise ValueError("Invalid deskew mode")



from typing import Tuple, Union

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.filters.kernels.butterworth import butterworth_kernel


def butterworth_filter(image,
                       shape=None,
                       cutoffs: Union[float, Tuple[float, ...]] = None,
                       cutoffs_in_freq_units=False,
                       epsilon: float = 1,
                       order: int = 3,
                       mode: str = 'reflect',
                       use_fft: bool = False,
                       internal_dtype=None):
    """
    Applies a Butterworth filter to an image.
    The Butterworth filter is a type of signal processing filter designed to have a frequency response
    as flat as possible in the passband. It is also referred to as a maximally flat magnitude filter.
    It was first described in 1930 by the British engineer and physicist Stephen Butterworth in his paper
    entitled "On the Theory of Filter Amplifiers". (source: wikipedia, https://en.wikipedia.org/wiki/Butterworth_filter)

    Parameters
    ----------
    image : image to apply filter to
    shape : filter shape
    cutoffs : Butterworth cutoffs.
    cutoffs_in_freq_units : If True, the cutoffs are specified in frequency units. If False, the units are in normalised within [0,1]
    order : Butterworth filter order
    mode : mode for convolution
    use_fft : True to use FFT
    internal_dtype : internal dtype used for computation

    Returns
    -------
    Filtered image.

    """
    sp = Backend.get_sp_module()

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype)

    if shape is None:
        shape = (11,) * image.ndim
    elif type(shape) is int and image.ndim > 1:
        shape = (shape,) * image.ndim

    if cutoffs is None:
        cutoffs = (0.5,) * image.ndim
    elif type(cutoffs) is float and image.ndim > 1:
        cutoffs = (cutoffs,) * image.ndim

    butterworth_filter = butterworth_kernel(shape=shape,
                                            cutoffs=cutoffs,
                                            cutoffs_in_freq_units=cutoffs_in_freq_units,
                                            epsilon=epsilon,
                                            order=order)

    image = Backend.to_backend(image)

    if use_fft:
        filtered_image = fft_convolve(image, butterworth_filter, mode=mode)
    else:
        filtered_image = sp.ndimage.convolve(image, butterworth_filter, mode=mode)

    filtered_image = filtered_image.astype(original_dtype, copy=False)

    return filtered_image



import os
import random
import tempfile
from os.path import join

import numpy
from arbol import aprint, asection
from dask.array.image import imread
from skimage.data import astronaut

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.video.crop_resize_pad import crop_resize_pad_image_sequence


def demo_crop_resize_pad_numpy():
    with NumpyBackend():
        demo_crop_resize_pad(display=True)


def demo_crop_resize_pad_cupy():
    try:
        with CupyBackend():
            demo_crop_resize_pad(display=True)
            return True

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def demo_crop_resize_pad(n=64, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("Prepare dataset..."):
        image = Backend.to_backend(astronaut()[0:500, 0:500])

        # generate reference 'ground truth' timelapse
        images = (image.copy() for _ in range(n))

        # modify each image:
        images = (sp.ndimage.shift(image, shift=(random.uniform(-1, 2), random.uniform(-2, 1), 0)) for image in images)

        # Convert back images to 8 bit:
        images = list(image.astype(numpy.uint8) for image in images)

    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        with asection("Save image sequences..."):
            # Create two subfolders for the two image sequences:
            folder = join(tmpdir, 'images')
            os.makedirs(folder)

            # Save images from sequence:
            for i, image in enumerate(images):
                from PIL import Image
                im = Image.fromarray(Backend.to_numpy(image))
                im.save(join(folder, f"frame_{i:05}.png"))

        with asection("resize sequence..."):
            # Create output folder:
            output_folder = join(tmpdir, 'resized')
            os.makedirs(output_folder)

            # Perform blend:
            crop_resize_pad_image_sequence(input_path=folder,
                                        output_path=output_folder,
                                        crop=3,
                                        resize=(500, 400),
                                        pad_width=(3, 7))

        # load images into dask arrays:
        images = imread(os.path.join(folder, 'frame_*.png'))
        images_resized = imread(os.path.join(output_folder, 'frame_*.png'))

        if display:
            from napari import Viewer, gui_qt
            with gui_qt():
                def _c(array):
                    return Backend.to_numpy(array)

                viewer = Viewer(ndisplay=2)
                viewer.add_image(_c(images), name='images_u', rgb=True)
                viewer.add_image(_c(images_resized), name='images_v', rgb=True)

                viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_crop_resize_pad_cupy():
        demo_crop_resize_pad_numpy()



from typing import Tuple, Sequence

import numpy
from dask.array import Array

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.processing.registration.sequence import image_stabilisation
from dexp.processing.utils.projection_generator import projection_generator


def image_stabilisation_proj(image: 'Array',
                             axis: int = 0,
                             projection_type: str = 'max-min',
                             **kwargs
                             ) -> SequenceRegistrationModel:
    """
    Computes a sequence stabilisation model for an image sequence indexed along a specified axis.
    Instead of running a full nD registration, this uses projections instead, economising memory and compute.
    Stabilises all 2D projections and then combines the information to figure out what is the best
    stabilisation for the whole nD image


    Parameters
    ----------
    image: image to stabilise
    axis: sequence axis along which to stabilise image
    projection_type : Projection type to use when in 'projection' mode: 'mean', 'min', 'max', 'max-min'
    kwargs: argument passthrough to the 'sequence_stabilisation' method.

    Returns
    -------
    sequence registration model

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    ndim = image.ndim
    image = xp.moveaxis(image, axis, 0)
    projections = projection_generator(image,
                                       axis_range=(1, ndim),
                                       projection_type=projection_type)

    return image_stabilisation_proj_(projections=projections,
                                     ndim=ndim - 1,
                                     **kwargs)


def image_stabilisation_proj_(projections: Sequence[Tuple],
                              ndim: int,
                              keep_best: bool = True,
                              debug_output: str = None,
                              **kwargs
                              ) -> SequenceRegistrationModel:
    """
    Same as 'sequence_stabilisation_proj' but takes projections instead, usefull if the projections are already available.

    Parameters
    ----------
    projections: List of projections as a list of tuples: (u, v, projection) where u and v are the axis indices and projection is an array where the first axis is the stabilisation axis.
    ndim: number of dimensions of original array
    keep_best: keep only the best n stabilised projections
    kwargs: argument passthrough to the 'sequence_stabilisation' method.

    Returns
    -------
    Sequence registration model

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # we stabilise along each projection:
    seq_reg_models = list((u - 1, v - 1, image_stabilisation(image=projection,
                                                             axis=0,
                                                             debug_output=f'{u - 1}_{v - 1}_{debug_output}' if debug_output is not None else debug_output,
                                                             **kwargs)) for u, v, projection in projections)

    # Let's sort the models by decreasing confidence:
    seq_reg_models = sorted(seq_reg_models, key=lambda t: t[2].overall_confidence(), reverse=True)

    if keep_best:
        # Only keep the best models, but enough to cover all coordinates:
        counts = numpy.zeros(shape=(ndim,), dtype=numpy.float32)
        selected_seq_reg_models = []
        for u, v, seq_reg_model in seq_reg_models:
            counts[u] += 1
            counts[v] += 1
            selected_seq_reg_models.append((u, v, seq_reg_model))
            if numpy.all(counts > 0):
                break
        seq_reg_models = selected_seq_reg_models

    # first we figure out the length of the sequence, and figure out the kind of model:
    _, _, model = seq_reg_models[0]
    length = len(model.model_list)

    if type(model.model_list[0]) == TranslationRegistrationModel:

        # Create a new model list:
        fused_model_list = [TranslationRegistrationModel(shift_vector=numpy.zeros(shape=(ndim,), dtype=numpy.float32)) for _ in range(length)]

        # Iterate through the sequence updating the fused model:
        for tp in range(length):

            # current fused model:
            fused_model = fused_model_list[tp]

            # keeps track of how many shift values we add here:
            counts = numpy.zeros(shape=(ndim,), dtype=numpy.float32)

            # We extract the models for the current time point (tp):
            models = list(tuple((u, v, seq_reg_model.model_list[tp])) for u, v, seq_reg_model in seq_reg_models)

            for u, v, model in models:
                model: TranslationRegistrationModel

                # we add the shifts at the right places:
                if not keep_best or counts[u] == 0:
                    fused_model.shift_vector[u] += model.shift_vector[0]
                    counts[u] += 1
                if not keep_best or counts[v] == 0:
                    fused_model.shift_vector[v] += model.shift_vector[1]
                    counts[v] += 1

            # ... so we can compute averages:
            fused_model.shift_vector /= counts

        model: SequenceRegistrationModel = SequenceRegistrationModel(model_list=fused_model_list)

        if debug_output is not None:
            model.plot(debug_output + '_fused_model')

        return model

    else:
        raise ValueError("Pairwise registration model not supported!")
